"""Qt worker that runs the regrid usecase."""
from __future__ import annotations

from PySide6.QtCore import QObject, Signal as pyqtSignal

from takefits.core import usecases
from takefits.core.app_state import AppState, create_app_state


class Regridder(QObject):
    """Backend worker that performs FITS regridding operations with Qt signals."""

    progress = pyqtSignal(int)
    finished = pyqtSignal(object, object)
    error = pyqtSignal(str)

    def __init__(
        self,
        original_data=None,
        original_wcs=None,
        original_header=None,
        filename=None,
        state: AppState | None = None,
    ):
        super().__init__()
        if state is None:
            self._state = create_app_state(
                data=original_data,
                header=original_header,
                wcs=original_wcs,
                filepath=filename,
                spectral_metadata={},
            )
        else:
            self._state = state

    def perform_regrid(self, params: dict):
        try:
            result = usecases.compute_regrid(
                self._state,
                params,
                progress_callback=self.progress.emit,
            )
        except Exception as exc:
            self.error.emit(str(exc))
            return None

        self.finished.emit(result.data, result.header)
        return result.data, result.header



# custom_colormap.py

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
import numpy as np

class CustomColormap:
    def __init__(self, name, cdict):
        self.name = name
        self.cdict = cdict
        self.colormap = LinearSegmentedColormap(name, cdict)
    
    def get_colormap(self):
        return self.colormap
    
    def plot_colormap(self):
        gradient = np.linspace(0, 1, 256)
        gradient = np.vstack((gradient, gradient))

        fig, ax = plt.subplots(figsize=(5, 2))
        ax.imshow(gradient, aspect='auto', cmap=self.colormap)
        ax.set_axis_off()
        plt.title(self.name)
        plt.show()

    def reversed_colormap(self):
        reversed_cdict = {color: [(1.0 - pos, val1, val2) for pos, val1, val2 in reversed(self.cdict[color])] for color in self.cdict}
        return CustomColormap(self.name + '_r', reversed_cdict)

    def apply_gamma(self, gamma_value):
        gamma_cmap = self.colormap(np.linspace(0, 1, self.colormap.N) ** gamma_value)
        return ListedColormap(gamma_cmap)

class ColorDefinitions:
    @staticmethod
    def rainbow():
        return {'red':   ((0.0, 0.0, 0.0),
                          (0.075, 0.25, 0.25),
                          (0.25, 0.0, 0.0),
                          (0.4, 0.0, 0.0),
                          (0.75, 1.0, 1.0),
                          (0.9, 1.0, 1.0),
                          (1.0, 1.0, 1.0)),
                'green': ((0.0, 0.0, 0.0),
                          (0.1, 0.0, 0.0),
                          (0.25, 0.0, 0.0),
                          (0.4, 0.75, 0.75),
                          (0.75, 1.0, 1.0),
                          (0.9, 0.0, 0.0),
                          (1.0, 1.0, 1.0)),
                'blue':  ((0.0, 0.0, 0.0),
                          (0.075, 0.25, 0.25),
                          (0.25, 0.8, 0.8),
                          (0.4, 1.0, 1.0),
                          (0.75, 0.0, 0.0),
                          (0.9, 0.0, 0.0),
                          (1.0, 1.0, 1.0))}

    @staticmethod
    def cool():
        return {'red':   ((0.0, 0.0, 0.0),
                          (0.75, 0.0, 0.0),
                          (1.0, 1.0, 1.0)),
                'green': ((0.0, 0.0, 0.0),
                          (0.4, 0.0, 0.0),
                          (0.7, 0.5, 0.5),
                          (0.95, 1.0, 1.0),
                          (1.0, 1.0, 1.0)),
                'blue':  ((0.0, 0.0, 0.0),
                          (0.5, 0.9, 0.9),
                          (1.0, 1.0, 1.0))}



from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QPushButton,
    QLineEdit, QMessageBox, QGroupBox, QListWidget
)
from PySide6.QtCore import Qt, Signal as pyqtSignal
from PySide6.QtGui import QFont
import numpy as np
import ast
from datetime import datetime
from takefits.ui.save_fits_dialog import SaveFITS
from takefits.core.history_provenance import build_processing_history_lines


from takefits.core.usecases import compute_arithmetic
from takefits.tools.base_panel import (
    capture_preferred_cursor_snapshot,
    clear_action_preview_record,
    confirm_pending_close,
    has_action_record_tag,
    record_action_preview,
    replay_action_history_to_current_cursor,
)

class UnitPreservationChecker(ast.NodeVisitor):
    """
    Checks if an arithmetic expression preserves the original unit (BUNIT).
    Returns True if the operation is unit-preserving (e.g., A + constant, A * constant),
    False if the unit changes (e.g., A * B, np.log10(A)).
    """
    # Functions that preserve units
    PRESERVING_FUNCTIONS = {
        'abs', 'absolute', 'fabs', 
        'max', 'min', 'amax', 'amin', 'nanmax', 'nanmin',
        'fmax', 'fmin', 'maximum', 'minimum',
        'mean', 'nanmean', 'median', 'nanmedian',
        'std', 'nanstd', 'ptp', 
        'sum', 'nansum', # Sum preserves unit string (e.g. K + K = K)
        'copysign', 'trunc', 'floor', 'ceil', 'rint'
    }

    def __init__(self, variable_names):
        self.variable_names = set(variable_names)
        self.preserves_unit = True
        self.uses_multiple_variables = False
        
    def visit_BinOp(self, node):
        # Check if both operands are variables (e.g., A * B)
        left_is_var = self._is_variable(node.left)
        right_is_var = self._is_variable(node.right)
        
        if left_is_var and right_is_var:
            # Two variables: unit changes (e.g., A * B, A / B)
            # Exception: Addition/Subtraction of same unit variables *should* preserve
            # But we don't know if units are same. Assuming they are for same-type operations?
            # Current logic conservative: A+B -> False.
            # Let's keep it conservative for multi-variable ops unless explicitly handled.
            self.preserves_unit = False
            self.uses_multiple_variables = True
        elif isinstance(node.op, (ast.Mult, ast.Div, ast.FloorDiv, ast.Pow, ast.Mod)):
            # Multiplication/Division/Power with variable always changes unit 
            # (unless multiplying by unitless constant, but strictly unit string might imply scale?)
            # Usually: K * 2 -> K. K * K -> K^2.
            # Here scalar * var -> preserves.
            if not (self._is_constant(node.left) or self._is_constant(node.right)):
                # If neither is constant (i.e. both vars), already caught above.
                # If one is constant:
                # A * 2 -> Preserves.
                pass
        elif isinstance(node.op, (ast.Add, ast.Sub)):
            # Addition or subtraction: preserve unit if one operand is constant
            if not (self._is_constant(node.left) or self._is_constant(node.right)):
                # A + B caught above.
                pass
        
        self.generic_visit(node)
    
    def visit_Call(self, node):
        # Function calls
        func_name = None
        if isinstance(node.func, ast.Attribute):
            # np.max, np.abs
            func_name = node.func.attr
        elif isinstance(node.func, ast.Name):
            # abs, max
            func_name = node.func.id
            
        if func_name not in self.PRESERVING_FUNCTIONS:
             self.preserves_unit = False
             
        self.generic_visit(node)
    
    def visit_UnaryOp(self, node):
        # Unary operations like -A, +A preserve units
        if isinstance(node.op, (ast.UAdd, ast.USub)):
            # +A, -A preserve unit
            pass
        else:
            # ~A (invert) changes unit or applies to mask
            self.preserves_unit = False
        self.generic_visit(node)
    
    def _is_variable(self, node):
        """Check if node is a variable (A, B, etc.)"""
        # Also handles Attributes if they are variables? No, variables are Names in this context.
        if isinstance(node, ast.Name):
            return node.id in self.variable_names
        return False
    
    def _is_constant(self, node):
        """Check if node is a constant (number)"""
        return isinstance(node, (ast.Constant, ast.Num)) or \
               (isinstance(node, ast.UnaryOp) and 
                isinstance(node.operand, (ast.Constant, ast.Num)))

class CubeArithmeticPanel(QWidget):
    """
    A tool for performing mathematical operations on FITS data cubes.
    Supports arithmetic, numpy functions, and masking.
    """
    
    destroyed = pyqtSignal()
    
    def __init__(self, fits_viewer, subwindows, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)
        self.fits_viewer = fits_viewer
        self.subwindows = subwindows
        self._action_record_tag = "panel:arithmetic"
        
        self.variables = {} 
        self.original_data_map = {} # Store original data for all windows
        self.original_bunit_map = {} # Store original BUNIT state for all windows
        self._has_restored_preview = False
        
        self.initUI()
        self.refresh_sources()
        self.resync_after_workspace_restore()

    def initUI(self):
        self.setWindowTitle("Arithmetic")
        self.resize(360, 220)
        
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(5, 5, 5, 5) # Tight margins
        main_layout.setSpacing(5)

        # 1. Variables List
        source_group = QGroupBox("Variables")
        source_layout = QVBoxLayout()
        source_layout.setContentsMargins(2, 5, 2, 2)
        source_layout.setSpacing(2)
        
        self.source_list = QListWidget()
        self.source_list.setFixedHeight(38) # Approx 2 lines, tighter
        self.source_list.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.source_list.setStyleSheet("font-size: 11px;")
        self.source_list.itemClicked.connect(self.handle_list_click)
        
        # Refresh Button (Right aligned, standard style)
        refresh_layout = QHBoxLayout()
        refresh_layout.addStretch()
        
        
        refresh_btn = QPushButton("Refresh")
        refresh_btn.setFixedSize(60, 24)
        refresh_btn.setStyleSheet("font-size: 10px; padding: 2px;")
        refresh_btn.clicked.connect(self.refresh_sources)
        
        refresh_layout.addWidget(refresh_btn)
        
        source_layout.addWidget(self.source_list)
        source_layout.addLayout(refresh_layout)
        source_group.setLayout(source_layout)
        main_layout.addWidget(source_group)

        # 2. Expression Input
        expr_group = QGroupBox("Formula (numpy syntax)")
        expr_layout = QVBoxLayout()
        expr_layout.setContentsMargins(2, 5, 2, 2)
        
        self.expr_input = QLineEdit()
        self.expr_input.setPlaceholderText("e.g. np.log10(A), np.where(B > 1, A/B, np.nan), etc.")
        self.expr_input.setAttribute(Qt.WidgetAttribute.WA_MacShowFocusRect, False)
        self.expr_input.setFont(QFont("Menlo", 12))
        self.expr_input.returnPressed.connect(self.compute_and_execute)
        
        expr_layout.addWidget(self.expr_input)
        expr_group.setLayout(expr_layout)
        main_layout.addWidget(expr_group)

        # 3. Actions (No grouping)
        
        # Preview and Reset Buttons
        preview_reset_layout = QHBoxLayout()
        preview_reset_layout.setSpacing(5)
        
        self.preview_btn = QPushButton("Preview Result")
        self.preview_btn.clicked.connect(self.apply_preview)
        
        self.reset_btn = QPushButton("Reset")
        self.reset_btn.clicked.connect(self.reset_preview)
        self.reset_btn.setFixedWidth(60)
        
        preview_reset_layout.addWidget(self.preview_btn)
        preview_reset_layout.addWidget(self.reset_btn)
        
        main_layout.addLayout(preview_reset_layout)
        
        self.save_btn = QPushButton("Save as FITS")
        self.save_btn.clicked.connect(self.save_result)
        
        main_layout.addWidget(self.save_btn)
        main_layout.addStretch()

    def show_message(self, message):
         QMessageBox.information(self, "Info", message)
         
    def _get_shape_str(self, data):
        """Returns string representation of shape (X, Y, Z)."""
        if data is None: return ""
        # numpy shape is (Z, Y, X) or (T, Z, Y, X)
        # We want to display (X, Y, Z)
        shape = data.shape
        if len(shape) >= 3:
            # Reverse the last 3 dimensions to get (X, Y, Z)
            dims = shape[-3:][::-1]
            return f"({', '.join(map(str, dims))})"
        else:
            # Fallback for 2D or 1D
            return str(shape[::-1])

    def refresh_sources(self):
        self.variables = {}
        self.source_list.clear()
        
        # A = Current Viewer Data
        if hasattr(self.fits_viewer, 'data'):
            data = self.fits_viewer.data
            shape_str = self._get_shape_str(data)
            self.variables["A"] = {
                "data": data, 
                "header": self.fits_viewer.header,
                "name": self.fits_viewer.filename
            }
            self.source_list.addItem(f"A: {self.fits_viewer.filename} {shape_str}")
            
        # B = Optional Second File
        if "B" not in self.variables:
             self.source_list.addItem("B: [Click to Load]")

    def handle_list_click(self, item):
        if item.text().startswith("B:"):
             self.load_file_b()

    def load_file_b(self):
        from PySide6.QtWidgets import QFileDialog
        path, _ = QFileDialog.getOpenFileName(self, "Load FITS for Variable B", "", "FITS Files (*.fits)", options=QFileDialog.Option.DontUseNativeDialog)
        if path:
            from astropy.io import fits
            import os
            try:
                with fits.open(path) as hdul:
                    data = hdul[0].data
                    header = hdul[0].header
                
                self.variables["B"] = {"data": data, "header": header, "name": os.path.basename(path)}
                self.variables["B"]["path"] = path
                # Update list item
                self.source_list.clear()
                
                # Re-add A
                a_info = self.variables.get('A', {})
                a_name = a_info.get('name', 'N/A')
                a_shape = self._get_shape_str(a_info.get('data'))
                self.source_list.addItem(f"A: {a_name} {a_shape}")
                
                 # Add B
                b_name = self.variables['B']['name']
                b_shape = self._get_shape_str(data)
                self.source_list.addItem(f"B: {b_name} {b_shape}")
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Failed to load file: {e}")

    def compute_and_execute(self):
        # Auto-activate preview on Enter
        self.apply_preview()

    def run_computation(self):
        expr_str = self.expr_input.text().strip()
        if not expr_str:
            return None, None

        # Prepare arguments for usecase
        data_a = self.variables.get("A", {}).get("data")
        header_a = self.variables.get("A", {}).get("header")
        
        if data_a is None:
            QMessageBox.warning(self, "Data Error", "Primary data (A) is not available.")
            return None, None
            
        data_b = self.variables.get("B", {}).get("data")

        try:
            # Delegate computation to core usecase
            result = compute_arithmetic(
                data_a=data_a,
                operation="expression", 
                data_b=data_b,
                expression=expr_str
            )
            return result, header_a
            
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Computation failed:\n{e}")
            return None, None

    def apply_preview(self):
        """Calculates and applies the result to headers/display."""
        data, header = self.run_computation()
        if data is not None:
            # Cache original data if not already cached
            if not self.original_data_map:
                self.original_data_map[self.fits_viewer] = self.fits_viewer.data
                for sub in self.subwindows:
                    if sub:
                        self.original_data_map[sub] = sub.data
            if not self.original_bunit_map:
                self._capture_original_bunit_map()

            # Update all windows
            viewer_data = self.original_data_map.get(self.fits_viewer)
            if viewer_data is not None and data.ndim != viewer_data.ndim:
                    QMessageBox.warning(self, "Warning", "Result dimension differs. Viewer update might fail.")

            self._apply_data_to_all(data)
            preserves_unit = self._expression_preserves_unit(self.expr_input.text().strip())
            self._apply_preview_bunit_state(preserves_unit)
            self._record_preview_action()
    
    def reset_preview(self):
        """Restores the original data from cache."""
        preferred_cursor = capture_preferred_cursor_snapshot(self.fits_viewer)
        removed_preview = self._clear_preview_action()
        restored_from_history = False
        if removed_preview:
            restored_from_history = replay_action_history_to_current_cursor(
                self.fits_viewer,
                preferred_cursor=preferred_cursor,
            )

        if not restored_from_history and self.original_data_map:
            # Restore each window
            for window, orig_data in self.original_data_map.items():
                if window: # Check if still alive
                        self._update_window_data(window, orig_data)
            
            self._sync_app_state_data()
            self.update_all_displays()
        self._restore_original_bunit_state()
        self._sync_bunit_to_viewers()
        self._sync_app_state_data()
        self._refresh_live_intensity_labels()
        self.original_data_map.clear()
        self.original_bunit_map.clear()
        self._has_restored_preview = False

    def _record_preview_action(self):
        expr_str = self.expr_input.text().strip()
        if not expr_str:
            return
        payload = {
            "operation": "expression",
            "expression": expr_str,
        }
        data_b_path = self.variables.get("B", {}).get("path")
        if data_b_path:
            payload["data_b_path"] = data_b_path
        record_action_preview(
            self.fits_viewer,
            "compute_arithmetic",
            payload,
            replace_tag=self._action_record_tag,
        )

    def _clear_preview_action(self) -> bool:
        return bool(
            clear_action_preview_record(
                self.fits_viewer,
                self._action_record_tag,
                action_name="compute_arithmetic",
            )
        )

    def _has_pending_preview_in_history(self) -> bool:
        return bool(has_action_record_tag(self.fits_viewer, self._action_record_tag))

    def resync_after_workspace_restore(self) -> None:
        self._has_restored_preview = self._has_pending_preview_in_history()

    def _apply_data_to_all(self, new_data):
        """Apply new data to all windows and refresh."""
        all_windows = [self.fits_viewer] + [s for s in self.subwindows if s]
        for window in all_windows:
            self._update_window_data(window, new_data)
        
        self._sync_app_state_data()
        self.update_all_displays()

    def _update_window_data(self, window, data):
        """Helper to update window.data and window.cube safely."""
        window.data = data
        if hasattr(window, 'update_cube'):
            window.update_cube()
            return
        if data.ndim > 2:
            if data.ndim == 3:
                window.cube = window.data
            elif data.ndim == 4:
                window.cube = window.data[0]

    def _sync_app_state_data(self):
        """Keep MainWindow.app_state data in sync with arithmetic preview/apply state."""
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        if hasattr(main_window, 'sync_app_state_data'):
            main_window.sync_app_state_data(
                data=self.fits_viewer.data,
                header=self.fits_viewer.header,
                wcs=self.fits_viewer.wcs,
            )

    def _capture_original_bunit_map(self):
        """Capture original per-window BUNIT state so preview/reset can toggle units safely."""
        self.original_bunit_map = {}
        all_windows = [self.fits_viewer] + [s for s in self.subwindows if s]
        for window in all_windows:
            header = getattr(window, 'header', None)
            if header is None:
                continue
            has_bunit = False
            bunit_value = None
            try:
                has_bunit = 'BUNIT' in header
                if has_bunit:
                    bunit_value = header.get('BUNIT')
            except Exception:
                has_bunit = False
                bunit_value = None
            self.original_bunit_map[window] = (has_bunit, bunit_value)

    def _restore_original_bunit_state(self):
        """Restore original BUNIT state captured before arithmetic preview."""
        if not self.original_bunit_map:
            return
        for window, state in self.original_bunit_map.items():
            header = getattr(window, 'header', None)
            if header is None:
                continue
            had_bunit, bunit_value = state
            try:
                if had_bunit:
                    header['BUNIT'] = bunit_value
                else:
                    header.pop('BUNIT', None)
            except Exception:
                continue

    def _apply_preview_bunit_state(self, preserves_unit: bool):
        """Apply preview-time BUNIT policy and refresh live labels."""
        if not self.original_bunit_map:
            self._capture_original_bunit_map()

        for window, state in self.original_bunit_map.items():
            header = getattr(window, 'header', None)
            if header is None:
                continue
            had_bunit, bunit_value = state
            try:
                if preserves_unit:
                    if had_bunit:
                        header['BUNIT'] = bunit_value
                    else:
                        header.pop('BUNIT', None)
                else:
                    header.pop('BUNIT', None)
            except Exception:
                continue

        self._sync_bunit_to_viewers()
        self._sync_app_state_data()
        self._refresh_live_intensity_labels()

    def _sync_bunit_to_viewers(self):
        """Keep live intensity unit labels synchronized with current header BUNIT."""
        header = getattr(self.fits_viewer, 'header', None)
        bunit = ""
        if header is not None:
            try:
                bunit = str(header.get('BUNIT', '') or '').strip()
            except Exception:
                bunit = ""

        try:
            self.fits_viewer.bunit = bunit
        except Exception:
            pass

        for window in self.subwindows:
            if not window:
                continue
            try:
                window.bunit = bunit
            except Exception:
                continue

    def _refresh_live_intensity_labels(self):
        """Refresh lower-right coordinate/intensity labels without requiring a new click."""
        main_window = getattr(self.fits_viewer, 'main_window', None) or self.fits_viewer
        refresh_fn = getattr(main_window, '_refresh_wcs_display_strings', None)
        if callable(refresh_fn):
            try:
                refresh_fn()
                return
            except Exception:
                pass
        try:
            self.fits_viewer.update_clicked_pix(
                int(self.fits_viewer._get_shared_xpix()),
                int(self.fits_viewer._get_shared_ypix()),
                update_slices=False,
                fast_blit=True,
            )
        except Exception:
            return

    def _expression_preserves_unit(self, expr_str: str) -> bool:
        """Return whether expression keeps original BUNIT semantics."""
        variable_names = list(self.variables.keys())
        if not expr_str or not variable_names:
            return True
        try:
            tree = ast.parse(expr_str, mode='eval')
            checker = UnitPreservationChecker(variable_names)
            checker.visit(tree)
            return checker.preserves_unit
        except Exception:
            # Conservative fallback: unknown expression is treated as unit-changing.
            return False

    def update_all_displays(self):
        """Refreshes the display of the main viewer and all subwindows."""
        all_windows = [self.fits_viewer] + [s for s in self.subwindows if s]
        
        for window in all_windows:
            # Logic borrowed from ScalingPanel / FITSViewer update
            current_channel = 0
            if hasattr(window, 'current_channel_index'):
                try:
                    current_channel = window.current_channel_index()
                except Exception:
                    current_channel = 0
            
            # We use update_channel to properly refresh the slice and overlay
            if hasattr(window, 'update_channel'):
                try:
                    # If 2D (ndim < 3), update_channel is a no-op (due to recent fix).
                    # So we must update the image artist manually for 2D.
                    if window.data.ndim < 3:
                        if hasattr(window, 'im'):
                            window.im.set_data(window.data)                            
                            window.canvas.draw_idle()
                    else:
                        window.update_channel(window.plane, current_channel)
                        
                except Exception as e:
                    print(f"Error updating display for window: {e}")
            elif hasattr(window, 'im'):
                 window.im.set_data(window.data)
                 window.canvas.draw_idle()

    def save_result(self):
        data, header = self.run_computation()
        if data is not None:
            # Copy header to avoid modifying the original
            if header is not None:
                new_header = header.copy()
            else:
                from astropy.io import fits
                # Create a minimal header if none exists
                new_header = fits.Header()
            
            # Check if the operation preserves units
            expr_str = self.expr_input.text().strip()
            preserves_unit = self._expression_preserves_unit(expr_str)
            
            # Handle BUNIT based on operation type
            original_bunit = new_header.get('BUNIT', None)
            if not preserves_unit:
                # Remove BUNIT for operations that change units
                if 'BUNIT' in new_header:
                    del new_header['BUNIT']

            # Update data range if finite values exist; otherwise remove stale values
            finite_mask = np.isfinite(data)
            if np.any(finite_mask):
                finite_data = data[finite_mask]
                new_header['DATAMIN'] = float(finite_data.min())
                new_header['DATAMAX'] = float(finite_data.max())
            else:
                new_header.pop('DATAMIN', None)
                new_header.pop('DATAMAX', None)
            
            for entry in build_processing_history_lines(self.fits_viewer):
                new_header.add_history(entry)
            
            saver = SaveFITS(data, new_header, self.fits_viewer.filename)
            saver.save(suffix="calc")

    def closeEvent(self, event):
        has_pending_preview = bool(
            self.original_data_map
            or self._has_restored_preview
            or self._has_pending_preview_in_history()
        )
        if has_pending_preview:
            choice = confirm_pending_close(
                self,
                "Close Arithmetic Panel",
                "There are unapplied arithmetic preview changes.",
            )
            if choice == "cancel":
                event.ignore()
                return
            if choice == "discard":
                self.reset_preview()
            else:
                self.original_data_map.clear()
                self.original_bunit_map.clear()
                self._has_restored_preview = False
        super().closeEvent(event)



"""Masking helpers for usecases."""
from __future__ import annotations

from typing import Dict, Literal, Optional

import numpy as np
from scipy import ndimage

from takefits.core.app_state import AppState
from takefits.core.io.save_fits import write_fits


MaskCondition = Literal["less_than", "greater_than"]
MomentMaskAlgorithm = Literal["smoothed_hysteresis", "moment_masking"]
MaskPolarity = Literal["emission", "absorption"]
MomentMaskPreset = Literal["faint", "normal", "strict"]
NoiseMethod = Literal["diff_mad", "mad", "std"]


MOMENT_MASK_PRESETS: Dict[str, Dict[str, float | int]] = {
    "faint": {
        "smooth_xy_pix": 1.5,
        "smooth_v_chan": 1.0,
        "seed_sigma": 3.0,
        "grow_sigma": 2.0,
        "clip_sigma": 4.0,
        "expand_xy_pix": 2,
        "expand_v_chan": 1,
        "min_channels": 2,
        "min_voxels": 0,
        "connectivity": 26,
    },
    "normal": {
        "smooth_xy_pix": 1.2,
        "smooth_v_chan": 0.8,
        "seed_sigma": 3.5,
        "grow_sigma": 2.5,
        "clip_sigma": 5.0,
        "expand_xy_pix": 1,
        "expand_v_chan": 1,
        "min_channels": 2,
        "min_voxels": 0,
        "connectivity": 26,
    },
    "strict": {
        "smooth_xy_pix": 0.8,
        "smooth_v_chan": 0.6,
        "seed_sigma": 4.5,
        "grow_sigma": 3.0,
        "clip_sigma": 6.0,
        "expand_xy_pix": 1,
        "expand_v_chan": 0,
        "min_channels": 3,
        "min_voxels": 0,
        "connectivity": 18,
    },
}


def get_moment_mask_preset(preset: str = "normal") -> Dict[str, float | int]:
    """
    Return a copy of preset parameters for automatic moment masking.

    Args:
        preset: Preset name ('faint', 'normal', 'strict')

    Returns:
        Parameter dictionary
    """
    preset_key = str(preset or "normal").strip().lower()
    if preset_key not in MOMENT_MASK_PRESETS:
        raise ValueError(
            f"Unknown moment mask preset '{preset}'. "
            f"Available: {sorted(MOMENT_MASK_PRESETS.keys())}"
        )
    return dict(MOMENT_MASK_PRESETS[preset_key])


def _robust_sigma(values: np.ndarray) -> float:
    """Estimate sigma using MAD (scaled for Gaussian noise)."""
    if values.size == 0:
        return float("nan")
    median = np.median(values)
    mad = np.median(np.abs(values - median))
    return float(1.4826 * mad)


def estimate_noise_sigma(
    data: np.ndarray,
    method: str = "diff_mad",
    spectral_axis: int = -3,
) -> float:
    """
    Estimate noise sigma from data using a robust method.

    Args:
        data: Input array
        method: 'diff_mad', 'mad', or 'std'
        spectral_axis: Spectral axis index for diff-MAD (default -3 for z in z,y,x)

    Returns:
        Estimated 1-sigma noise level
    """
    array = np.asarray(data)
    finite = np.isfinite(array)
    if not np.any(finite):
        raise ValueError("No finite values available for noise estimation")

    method_key = str(method or "diff_mad").strip().lower()
    if method_key == "diff_mad":
        axis = spectral_axis if spectral_axis >= 0 else (array.ndim + spectral_axis)
        if 0 <= axis < array.ndim and array.shape[axis] > 1:
            diff = np.diff(array, axis=axis)
            diff_values = diff[np.isfinite(diff)]
            if diff_values.size > 0:
                sigma = _robust_sigma(diff_values) / np.sqrt(2.0)
                if np.isfinite(sigma) and sigma > 0:
                    return float(sigma)
        method_key = "mad"

    values = array[finite]
    if method_key == "mad":
        sigma = _robust_sigma(values)
        if np.isfinite(sigma) and sigma > 0:
            return float(sigma)
        method_key = "std"

    if method_key == "std":
        sigma = float(np.std(values))
        if np.isfinite(sigma) and sigma > 0:
            return sigma

    raise ValueError("Could not estimate a valid noise sigma")


def _resolve_connectivity_rank(connectivity: int, ndim: int) -> int:
    if ndim == 2:
        return 1 if connectivity <= 4 else 2
    if ndim == 3:
        if connectivity <= 6:
            return 1
        if connectivity <= 18:
            return 2
        return 3
    return 1


def _make_connectivity_structure(ndim: int, connectivity: int) -> np.ndarray:
    rank = _resolve_connectivity_rank(int(connectivity), ndim)
    return ndimage.generate_binary_structure(ndim, rank)


def _make_expansion_structure(expand_xy_pix: int, expand_v_chan: int) -> np.ndarray:
    xy = max(0, int(expand_xy_pix))
    vel = max(0, int(expand_v_chan))
    return np.ones((2 * vel + 1, 2 * xy + 1, 2 * xy + 1), dtype=bool)


def _filter_connected_components(
    mask: np.ndarray,
    min_voxels: int,
    min_channels: int,
    connectivity: int,
) -> np.ndarray:
    if not np.any(mask):
        return mask

    min_voxels_int = max(0, int(min_voxels))
    min_channels_int = max(1, int(min_channels))
    if min_voxels_int <= 0 and min_channels_int <= 1:
        return mask

    structure = _make_connectivity_structure(mask.ndim, connectivity)
    labels, n_labels = ndimage.label(mask, structure=structure)
    if n_labels == 0:
        return mask

    # Count voxels per component in C/NumPy instead of repeated np.where scans.
    counts = np.bincount(labels.ravel(), minlength=n_labels + 1)
    keep_components = np.ones(n_labels + 1, dtype=bool)
    keep_components[0] = False
    if min_voxels_int > 0:
        keep_components &= counts >= min_voxels_int
        keep_components[0] = False

    # Spectral extent check using find_objects (bounding boxes), cheaper than per-label indexing.
    if min_channels_int > 1 and mask.ndim >= 3:
        objects = ndimage.find_objects(labels)
        spectral_extent = np.zeros(n_labels + 1, dtype=np.int32)
        for label_index, slc in enumerate(objects, start=1):
            if slc is None:
                continue
            spectral_extent[label_index] = int(slc[0].stop) - int(slc[0].start)
        keep_components &= spectral_extent >= min_channels_int
        keep_components[0] = False

    return keep_components[labels]


def _resolve_moment_mask_params(
    preset: str,
    smooth_xy_pix: Optional[float],
    smooth_v_chan: Optional[float],
    seed_sigma: Optional[float],
    grow_sigma: Optional[float],
    clip_sigma: Optional[float],
    expand_xy_pix: Optional[int],
    expand_v_chan: Optional[int],
    min_channels: Optional[int],
    min_voxels: Optional[int],
    connectivity: Optional[int],
) -> Dict[str, float | int]:
    params = get_moment_mask_preset(preset)

    if smooth_xy_pix is not None:
        params["smooth_xy_pix"] = float(smooth_xy_pix)
    if smooth_v_chan is not None:
        params["smooth_v_chan"] = float(smooth_v_chan)
    if seed_sigma is not None:
        params["seed_sigma"] = float(seed_sigma)
    if grow_sigma is not None:
        params["grow_sigma"] = float(grow_sigma)
    if clip_sigma is not None:
        params["clip_sigma"] = float(clip_sigma)
    if expand_xy_pix is not None:
        params["expand_xy_pix"] = int(expand_xy_pix)
    if expand_v_chan is not None:
        params["expand_v_chan"] = int(expand_v_chan)
    if min_channels is not None:
        params["min_channels"] = int(min_channels)
    if min_voxels is not None:
        params["min_voxels"] = int(min_voxels)
    if connectivity is not None:
        params["connectivity"] = int(connectivity)

    params["smooth_xy_pix"] = max(0.0, float(params["smooth_xy_pix"]))
    params["smooth_v_chan"] = max(0.0, float(params["smooth_v_chan"]))
    params["seed_sigma"] = max(0.1, float(params["seed_sigma"]))
    params["grow_sigma"] = max(0.1, float(params["grow_sigma"]))
    params["clip_sigma"] = max(0.1, float(params["clip_sigma"]))
    params["expand_xy_pix"] = max(0, int(params["expand_xy_pix"]))
    params["expand_v_chan"] = max(0, int(params["expand_v_chan"]))
    params["min_channels"] = max(1, int(params["min_channels"]))
    params["min_voxels"] = max(0, int(params["min_voxels"]))
    params["connectivity"] = int(params["connectivity"])

    return params


def _smooth_for_moment_mask(
    cube: np.ndarray,
    smooth_xy_pix: float,
    smooth_v_chan: float,
    finite: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Fast Gaussian smoothing specialized for moment-mask generation.

    Uses scipy.ndimage directly and performs weighted smoothing when NaNs exist:
      smoothed = gaussian(data * valid) / gaussian(valid)
    """
    sigma = (
        max(0.0, float(smooth_v_chan)),
        max(0.0, float(smooth_xy_pix)),
        max(0.0, float(smooth_xy_pix)),
    )
    if sigma[0] == 0.0 and sigma[1] == 0.0 and sigma[2] == 0.0:
        return cube.copy()

    finite_mask = np.isfinite(cube) if finite is None else finite
    if np.all(finite_mask):
        return ndimage.gaussian_filter(cube, sigma=sigma, mode="reflect")

    cube_filled = np.where(finite_mask, cube, 0.0)
    valid = finite_mask.astype(np.float32, copy=False)
    smoothed_num = ndimage.gaussian_filter(cube_filled, sigma=sigma, mode="reflect")
    smoothed_den = ndimage.gaussian_filter(valid, sigma=sigma, mode="reflect")

    smoothed = np.empty_like(smoothed_num, dtype=smoothed_num.dtype)
    eps = np.finfo(smoothed_den.dtype).eps
    np.divide(smoothed_num, smoothed_den, out=smoothed, where=smoothed_den > eps)
    smoothed[smoothed_den <= eps] = np.nan
    return smoothed


def _compute_moment_mask_cube(
    cube: np.ndarray,
    algorithm: str,
    polarity: str,
    noise_method: str,
    params: Dict[str, float | int],
) -> np.ndarray:
    if cube.ndim != 3:
        raise ValueError(f"Expected 3D cube, got ndim={cube.ndim}")

    finite = np.isfinite(cube)
    if not np.any(finite):
        return np.zeros_like(cube, dtype=bool)

    polarity_key = str(polarity or "emission").strip().lower()
    if polarity_key not in ("emission", "absorption"):
        raise ValueError("polarity must be 'emission' or 'absorption'")

    working = cube if polarity_key == "emission" else -cube
    smooth_xy_pix = float(params["smooth_xy_pix"])
    smooth_v_chan = float(params["smooth_v_chan"])

    smoothed = _smooth_for_moment_mask(
        working,
        smooth_xy_pix=smooth_xy_pix,
        smooth_v_chan=smooth_v_chan,
        finite=finite,
    )

    sigma_original = estimate_noise_sigma(working, method=noise_method, spectral_axis=0)
    sigma_smoothed = estimate_noise_sigma(smoothed, method=noise_method, spectral_axis=0)

    eps = np.finfo(float).eps
    sigma_original = max(float(sigma_original), eps)
    sigma_smoothed = max(float(sigma_smoothed), eps)

    algorithm_key = str(algorithm or "smoothed_hysteresis").strip().lower()
    connectivity = int(params["connectivity"])

    if algorithm_key == "smoothed_hysteresis":
        seed_sigma = float(params["seed_sigma"])
        grow_sigma = float(params["grow_sigma"])

        seed_mask = smoothed >= (seed_sigma * sigma_smoothed)
        grow_mask = working >= (grow_sigma * sigma_original)
        candidate_mask = np.logical_or(seed_mask, grow_mask) & finite
        structure = _make_connectivity_structure(3, connectivity)
        mask = ndimage.binary_propagation(seed_mask, structure=structure, mask=candidate_mask)
        mask = mask & finite
    elif algorithm_key == "moment_masking":
        clip_sigma = float(params["clip_sigma"])
        expand_xy_pix = int(params["expand_xy_pix"])
        expand_v_chan = int(params["expand_v_chan"])

        seed_mask = smoothed >= (clip_sigma * sigma_smoothed)
        if expand_xy_pix > 0 or expand_v_chan > 0:
            expansion_structure = _make_expansion_structure(expand_xy_pix, expand_v_chan)
            mask = ndimage.binary_dilation(seed_mask, structure=expansion_structure)
        else:
            mask = seed_mask
        mask = mask & finite
    else:
        raise ValueError(
            f"Unknown moment mask algorithm '{algorithm}'. "
            "Expected 'smoothed_hysteresis' or 'moment_masking'."
        )

    return _filter_connected_components(
        mask.astype(bool),
        min_voxels=int(params["min_voxels"]),
        min_channels=int(params["min_channels"]),
        connectivity=connectivity,
    )


def compute_moment_mask(
    data: np.ndarray,
    algorithm: str = "smoothed_hysteresis",
    polarity: str = "emission",
    preset: str = "normal",
    smooth_xy_pix: Optional[float] = None,
    smooth_v_chan: Optional[float] = None,
    seed_sigma: Optional[float] = None,
    grow_sigma: Optional[float] = None,
    clip_sigma: Optional[float] = None,
    expand_xy_pix: Optional[int] = None,
    expand_v_chan: Optional[int] = None,
    min_channels: Optional[int] = None,
    min_voxels: Optional[int] = None,
    connectivity: Optional[int] = None,
    noise_method: str = "diff_mad",
) -> np.ndarray:
    """
    Create an automatic moment-analysis mask from a data cube.

    Supports two standard recipes:
      - smoothed_hysteresis: seed+grow mask from smoothed/original S/N thresholds
      - moment_masking: clip in smoothed cube, then expand back to original cube

    Args:
        data: Input 3D cube (z, y, x) or 4D cube (s, z, y, x)
        algorithm: 'smoothed_hysteresis' or 'moment_masking'
        polarity: 'emission' (positive lines) or 'absorption' (negative lines)
        preset: 'faint', 'normal', or 'strict'
        smooth_xy_pix, smooth_v_chan: Gaussian smoothing sigma in pixels/channels
        seed_sigma, grow_sigma: Thresholds for hysteresis recipe
        clip_sigma: Threshold for moment_masking recipe
        expand_xy_pix, expand_v_chan: Expansion radius for moment_masking recipe
        min_channels: Minimum spectral extent for connected components
        min_voxels: Minimum voxel count for connected components
        connectivity: Component connectivity (3D: 6, 18, 26)
        noise_method: 'diff_mad', 'mad', or 'std'

    Returns:
        Boolean mask with the same shape as data (True=keep, False=mask)
    """
    array = np.asarray(data)
    if array.ndim not in (3, 4):
        raise ValueError("compute_moment_mask expects 3D or 4D data")

    params = _resolve_moment_mask_params(
        preset=preset,
        smooth_xy_pix=smooth_xy_pix,
        smooth_v_chan=smooth_v_chan,
        seed_sigma=seed_sigma,
        grow_sigma=grow_sigma,
        clip_sigma=clip_sigma,
        expand_xy_pix=expand_xy_pix,
        expand_v_chan=expand_v_chan,
        min_channels=min_channels,
        min_voxels=min_voxels,
        connectivity=connectivity,
    )

    if array.ndim == 3:
        return _compute_moment_mask_cube(array, algorithm, polarity, noise_method, params)

    result = np.zeros_like(array, dtype=bool)
    for stokes_index in range(array.shape[0]):
        result[stokes_index] = _compute_moment_mask_cube(
            array[stokes_index],
            algorithm,
            polarity,
            noise_method,
            params,
        )
    return result


def apply_mask_moment_recipe(
    state: AppState,
    algorithm: str = "smoothed_hysteresis",
    polarity: str = "emission",
    preset: str = "normal",
    smooth_xy_pix: Optional[float] = None,
    smooth_v_chan: Optional[float] = None,
    seed_sigma: Optional[float] = None,
    grow_sigma: Optional[float] = None,
    clip_sigma: Optional[float] = None,
    expand_xy_pix: Optional[int] = None,
    expand_v_chan: Optional[int] = None,
    min_channels: Optional[int] = None,
    min_voxels: Optional[int] = None,
    connectivity: Optional[int] = None,
    noise_method: str = "diff_mad",
) -> AppState:
    """
    Apply automatic moment-analysis mask to state.data in-place.

    Args:
        state: The AppState to update
        algorithm: 'smoothed_hysteresis' or 'moment_masking'
        polarity: 'emission' or 'absorption'
        preset: 'faint', 'normal', or 'strict'
        smooth_xy_pix, smooth_v_chan: Optional smoothing overrides
        seed_sigma, grow_sigma, clip_sigma: Optional threshold overrides
        expand_xy_pix, expand_v_chan: Optional expansion overrides
        min_channels, min_voxels, connectivity: Component filters
        noise_method: 'diff_mad', 'mad', or 'std'

    Returns:
        Updated AppState
    """
    if state.data is None:
        raise ValueError("No data loaded")

    mask = compute_moment_mask(
        state.data,
        algorithm=algorithm,
        polarity=polarity,
        preset=preset,
        smooth_xy_pix=smooth_xy_pix,
        smooth_v_chan=smooth_v_chan,
        seed_sigma=seed_sigma,
        grow_sigma=grow_sigma,
        clip_sigma=clip_sigma,
        expand_xy_pix=expand_xy_pix,
        expand_v_chan=expand_v_chan,
        min_channels=min_channels,
        min_voxels=min_voxels,
        connectivity=connectivity,
        noise_method=noise_method,
    )

    masked = state.data.copy()
    if not np.issubdtype(masked.dtype, np.floating):
        masked = masked.astype(np.float32, copy=False)
    masked[~mask] = np.nan
    state.data = masked
    return state


def compute_masked(
    data: np.ndarray,
    threshold: float,
    condition: str = "less_than"
) -> np.ndarray:
    """
    Apply threshold mask to data (pure function).

    Pixels meeting the condition are set to NaN.

    Args:
        data: Input data array
        threshold: Threshold value
        condition: "less_than" or "greater_than"

    Returns:
        Masked data array (copy with NaN where condition met)
    """
    result = data.copy()

    if condition == "less_than":
        result[result < threshold] = np.nan
    elif condition == "greater_than":
        result[result > threshold] = np.nan
    else:
        raise ValueError(f"Unknown condition: {condition}")

    return result


def apply_mask_threshold(
    state: AppState,
    threshold: float,
    condition: str = "less_than"
) -> AppState:
    """
    Apply threshold mask to state.data in-place.

    Args:
        state: The AppState to update
        threshold: Threshold value
        condition: "less_than" or "greater_than"

    Returns:
        The updated AppState
    """
    if state.data is None:
        raise ValueError("No data loaded")

    state.data = compute_masked(state.data, threshold, condition)
    return state


def apply_mask_external(
    state: AppState,
    mask_path: str,
    mask_value: float = 0.0
) -> AppState:
    """
    Apply external FITS mask to state.data in-place.

    Pixels where mask == mask_value are set to NaN.

    Args:
        state: The AppState to update
        mask_path: Path to the mask FITS file
        mask_value: Value in mask that indicates masked pixels (default 0.0)

    Returns:
        The updated AppState

    Raises:
        ValueError: If mask shape doesn't match data shape
    """
    from astropy.io import fits

    if state.data is None:
        raise ValueError("No data loaded")

    with fits.open(mask_path) as hdul:
        mask_data = hdul[0].data

    if mask_data.shape != state.data.shape:
        raise ValueError(
            f"Mask shape {mask_data.shape} doesn't match data shape {state.data.shape}"
        )

    # Apply mask
    state.data = state.data.copy()
    state.data[mask_data == mask_value] = np.nan

    return state


def export_mask_fits(
    state: AppState,
    mask_data: np.ndarray,
    output_path: str,
    threshold: Optional[float] = None,
    condition: Optional[str] = None,
    history_entries: Optional[list] = None,
    mask_as_float: bool = False,
    nan_for_mask: bool = False,
) -> str:
    """
    Export a mask array to FITS file.

    Args:
        state: AppState with original header info
        mask_data: Mask array (1=unmasked, 0=masked)
        output_path: Path for output FITS file
        threshold: Optional threshold value (for HISTORY)
        condition: Optional condition (for HISTORY)
        history_entries: Optional list of HISTORY entries
        mask_as_float: Save mask as float32 instead of int16
        nan_for_mask: If saving as float, convert masked pixels (0) to NaN

    Returns:
        The output file path
    """
    from astropy.io import fits

    header = state.header.copy() if state.header is not None else fits.Header()

    # Remove header keys that conflict with mask data typing/scaling.
    original_naxis = int(header.get('NAXIS', 0) or 0)
    for key in ('SIMPLE', 'BITPIX', 'NAXIS', 'EXTEND', 'BSCALE', 'BZERO', 'BLANK', 'BUNIT'):
        if key in header:
            del header[key]
    for i in range(1, original_naxis + 1):
        key = f'NAXIS{i}'
        if key in header:
            del header[key]

    data_to_save = mask_data
    if mask_as_float:
        data_to_save = mask_data.astype(np.float32)
        if nan_for_mask:
            data_to_save = data_to_save.copy()
            data_to_save[data_to_save == 0] = np.nan

    # Update shape info
    header['NAXIS'] = data_to_save.ndim
    for i, dim in enumerate(data_to_save.shape[::-1], 1):
        header[f'NAXIS{i}'] = dim

    # Add masking info to header
    if threshold is not None:
        header['MASKTHRS'] = (threshold, 'Mask threshold value')
    if condition is not None:
        header['MASKCOND'] = (condition, 'Mask condition')

    # Add history
    if history_entries:
        for entry in history_entries:
            header.add_history(entry)

    if mask_as_float:
        data_to_save = data_to_save.astype(np.float32)
    else:
        data_to_save = data_to_save.astype(np.int16)

    return write_fits(output_path, data_to_save, header)



from PySide6.QtWidgets import QMenu, QMessageBox
from PySide6.QtGui import QAction, QActionGroup, QKeySequence
from PySide6.QtCore import Qt
from takefits.ui.config_panel import ConfigPanel
from takefits.logic.show_header import ShowHeader
from takefits.core.version import APP_NAME, APP_VERSION
from takefits.core.wcs_frames import (
    available_display_frames,
    display_frame_label,
    normalize_display_frame,
    preferred_display_frame,
)


class MenuBar:
    _instance = None 
    @staticmethod
    def get_instance(parent):
        if MenuBar._instance is None:
            MenuBar._instance = MenuBar(parent)
        return MenuBar._instance

    def __init__(self, parent):
        self.parent = parent
        self.make_menu()
        
    def make_menu(self):
        menubar = self.parent.menuBar()

        # File Menu
        file_menu = menubar.addMenu("File")
        self.save_workspace_action = QAction("Save Workspace", self.parent)
        self.load_workspace_action = QAction("Load Workspace...", self.parent)
        self.save_recipe_action = QAction("Save Recipe...", self.parent)
        self.load_recipe_action = QAction("Load Recipe...", self.parent)
        self.save_workspace_action.setShortcut(QKeySequence.StandardKey.Save)
        self.load_workspace_action.setShortcut(QKeySequence.StandardKey.Open)
        self.save_recipe_action.setShortcut(QKeySequence("Ctrl+Shift+S"))
        self.load_recipe_action.setShortcut(QKeySequence("Ctrl+Shift+O"))
        for action in (
            self.save_workspace_action,
            self.load_workspace_action,
            self.save_recipe_action,
            self.load_recipe_action,
        ):
            action.setShortcutContext(Qt.ShortcutContext.ApplicationShortcut)
        self.save_workspace_action.triggered.connect(self.save_workspace)
        self.load_workspace_action.triggered.connect(self.load_workspace)
        self.save_recipe_action.triggered.connect(self.save_recipe)
        self.load_recipe_action.triggered.connect(self.load_recipe)
        file_menu.addAction(self.save_workspace_action)
        file_menu.addAction(self.load_workspace_action)
        file_menu.addSeparator()
        file_menu.addAction(self.save_recipe_action)
        file_menu.addAction(self.load_recipe_action)
        file_menu.addSeparator()
        self.show_header_action = QAction("Show Header", self.parent)
        self.show_header_action.triggered.connect(self.open_header_panel)
        file_menu.addAction(self.show_header_action)

        # Actions Menu (avoid macOS auto-injected Edit entries like Writing Tools / Emoji & Symbols)
        edit_menu = menubar.addMenu("Actions")
        self.undo_action = QAction("Undo Analysis", self.parent)
        self.redo_action = QAction("Redo Analysis", self.parent)
        self.undo_action.setShortcut(QKeySequence("Alt+Left"))
        self.redo_action.setShortcut(QKeySequence("Alt+Right"))
        self.undo_action.triggered.connect(self.undo_action_triggered)
        self.redo_action.triggered.connect(self.redo_action_triggered)
        edit_menu.addAction(self.undo_action)
        edit_menu.addAction(self.redo_action)
        self.set_undo_redo_enabled(False, False)
        edit_menu.addSeparator()
        self.view_back_action = QAction("View Back", self.parent)
        self.view_forward_action = QAction("View Forward", self.parent)
        self.view_back_action.setShortcut(QKeySequence.StandardKey.Undo)
        self.view_forward_action.setShortcut(QKeySequence.StandardKey.Redo)
        self.view_back_action.triggered.connect(self.view_back_triggered)
        self.view_forward_action.triggered.connect(self.view_forward_triggered)
        edit_menu.addAction(self.view_back_action)
        edit_menu.addAction(self.view_forward_action)
        self.set_view_navigation_enabled(False, False)
        
        # Preferences
        dummy_menu = menubar.addMenu("Preferences")
        preferences_action = QAction("Preferences", self.parent)
        dummy_menu.addAction(preferences_action)
        preferences_action.triggered.connect(self.open_config_panel)
        

        # Window Menu
        window_menu = menubar.addMenu("Window")

        self.control_panel_action = QAction("ToolsPanel", self.parent, checkable=True)
        window_menu.addAction(self.control_panel_action)
        self.control_panel_action.triggered.connect(self.toggle_control_panel)
        self.control_panel_action.setChecked(True)
        
        self.range_panel_action = QAction("RangePanel", self.parent, checkable=True)
        window_menu.addAction(self.range_panel_action)
        self.range_panel_action.triggered.connect(self.toggle_range_panel)
        self.range_panel_action.setChecked(True)
        
        window_menu.addSeparator()
        self.main_action = QAction("X-Y (Main)", self.parent, checkable=True)
        self.sub1_action = QAction("X-Z (Sub1)", self.parent, checkable=True)
        self.sub2_action = QAction("Z-Y (Sub2)", self.parent, checkable=True)
        
        self.main_action.setChecked(True)
        
        self.main_action.triggered.connect(self.toggle_main_window)
        self.sub1_action.triggered.connect(self.toggle_sub1_window)
        self.sub2_action.triggered.connect(self.toggle_sub2_window)
        
        window_menu.addAction(self.main_action)
        window_menu.addAction(self.sub1_action)
        window_menu.addAction(self.sub2_action)

        wcs_menu = menubar.addMenu("WCS")
        self.wcs_frame_group = QActionGroup(self.parent)
        self.wcs_frame_group.setExclusive(True)
        self.wcs_frame_actions = {}
        for frame in available_display_frames(getattr(self.parent, "wcs", None)):
            action = QAction(display_frame_label(frame), self.parent, checkable=True)
            action.triggered.connect(lambda checked=False, frame_name=frame: self.set_wcs_frame(frame_name))
            self.wcs_frame_group.addAction(action)
            wcs_menu.addAction(action)
            self.wcs_frame_actions[frame] = action
        if self.wcs_frame_actions:
            wcs_menu.addSeparator()
        self.wcs_decimal_group = QActionGroup(self.parent)
        self.wcs_decimal_group.setExclusive(True)
        self.wcs_decimal_action = QAction("Decimal", self.parent, checkable=True)
        self.wcs_sexagesimal_action = QAction("Sexagesimal", self.parent, checkable=True)
        self.wcs_decimal_action.triggered.connect(lambda checked=False: self.set_wcs_decimal_mode(True))
        self.wcs_sexagesimal_action.triggered.connect(lambda checked=False: self.set_wcs_decimal_mode(False))
        self.wcs_decimal_group.addAction(self.wcs_decimal_action)
        self.wcs_decimal_group.addAction(self.wcs_sexagesimal_action)
        wcs_menu.addAction(self.wcs_decimal_action)
        wcs_menu.addAction(self.wcs_sexagesimal_action)

        current_frame = preferred_display_frame(getattr(self.parent, "wcs", None))
        getter = getattr(self.parent, "get_wcs_display_frame", None)
        if callable(getter):
            try:
                current_frame = normalize_display_frame(getter())
            except Exception:
                current_frame = preferred_display_frame(getattr(self.parent, "wcs", None))
        self.set_wcs_frame_checked(current_frame)
        decimal_mode = True
        decimal_getter = getattr(self.parent, "get_wcs_decimal_mode", None)
        if callable(decimal_getter):
            try:
                decimal_mode = bool(decimal_getter())
            except Exception:
                decimal_mode = True
        self.set_wcs_decimal_checked(decimal_mode)
        
        # Tools Menu
        tools_menu = menubar.addMenu("Tools")
        tools_menu.addSeparator() # Add a visual separator in the menu

        actions = []

        chmap_action = QAction("Channel Map", self.parent)
        chmap_action.triggered.connect(self.open_chmap_panel)
        if self.parent.wcs.wcs.naxis < 3: chmap_action.setEnabled(False)
        actions.append(chmap_action)

        clump_finding_action = QAction("Clump Finding", self.parent)
        clump_finding_action.triggered.connect(self.open_clump_finding_panel)
        actions.append(clump_finding_action)

        contour_action = QAction("Contours", self.parent)
        contour_action.triggered.connect(self.open_contour_panel)
        actions.append(contour_action)

        marker_action = QAction("Markers", self.parent)
        marker_action.triggered.connect(self.parent.open_marker_panel)
        actions.append(marker_action)

        colorscale_action = QAction("Color Settings", self.parent)
        colorscale_action.triggered.connect(self.open_colorscale_panel)
        actions.append(colorscale_action)
        
        arithmetic_action = QAction("Arithmetic", self.parent)
        arithmetic_action.triggered.connect(self.open_arithmetic_panel)
        actions.append(arithmetic_action)

        baseline_action = QAction("Baseline", self.parent)
        baseline_action.triggered.connect(self.open_baseline_panel)
        if self.parent.wcs.wcs.naxis < 3:
            baseline_action.setEnabled(False)
        actions.append(baseline_action)

        cutout_action = QAction("Cut Out", self.parent)
        cutout_action.triggered.connect(self.open_cutout_dialog)
        actions.append(cutout_action)

        integ_action = QAction("Integration", self.parent)
        integ_action.triggered.connect(self.open_integ_panel)
        if self.parent.wcs.wcs.naxis < 3:
            integ_action.setEnabled(False)
        actions.append(integ_action)

        mask_panel_action = QAction("Mask", self.parent)
        mask_panel_action.triggered.connect(self.open_mask_panel)
        actions.append(mask_panel_action)

        pvd_action = QAction("PV diagram", self.parent)
        pvd_action.triggered.connect(self.open_pvd_panel)
        if self.parent.wcs.wcs.naxis < 3:
            pvd_action.setEnabled(False)
        actions.append(pvd_action)

        regrid_action = QAction("Regrid", self.parent)
        regrid_action.triggered.connect(self.open_regrid_panel)
        actions.append(regrid_action)

        scaling_action = QAction("Scaling", self.parent)
        scaling_action.triggered.connect(self.open_scaling_panel)
        actions.append(scaling_action)

        smooth_action = QAction("Smoothing", self.parent)
        smooth_action.triggered.connect(self.open_smooth_panel)
        actions.append(smooth_action)

        spec_action = QAction("Spectrum", self.parent)
        spec_action.triggered.connect(self.open_spec_window)
        if self.parent.wcs.wcs.naxis < 3:
            spec_action.setEnabled(False)
        actions.append(spec_action)

        unit_conversion_action = QAction("Unit Conversion", self.parent)
        unit_conversion_action.triggered.connect(self.open_unit_conversion_panel)
        actions.append(unit_conversion_action)

        for action in sorted(actions, key=lambda a: a.text()):
            tools_menu.addAction(action)


        region_menu = menubar.addMenu("Region")
        self.circle_action = QAction("Circle", self.parent, checkable=True)
        self.rectangle_action = QAction("Rectangle", self.parent, checkable=True)
        self.ellipse_action = QAction("Ellipse", self.parent, checkable=True)
        self.cube_action = QAction("Cube", self.parent, checkable=True) 
        
        region_group = QActionGroup(self.parent)
        region_group.setExclusive(True)
        region_group.addAction(self.circle_action)
        region_group.addAction(self.rectangle_action)
        region_group.addAction(self.ellipse_action)
        region_group.addAction(self.cube_action) 
        
        region_menu.addAction(self.circle_action)
        region_menu.addAction(self.rectangle_action)
        region_menu.addAction(self.ellipse_action)
        region_menu.addAction(self.cube_action)

        if self.parent.wcs.wcs.naxis < 3:
            self.cube_action.setEnabled(False)

        region_menu.addSeparator()
        clear_all_action = QAction("Clear All", self.parent)
        clear_all_action.triggered.connect(self.parent.clear_all_regions_globally)
        region_menu.addAction(clear_all_action)
        region_menu.addSeparator()
        self.save_regions_action = QAction("Save Regions...", self.parent)
        self.save_regions_action.triggered.connect(self.parent.save_regions_dialog)
        region_menu.addAction(self.save_regions_action)
        self.load_regions_action = QAction("Load Regions...", self.parent)
        self.load_regions_action.triggered.connect(self.parent.load_regions_dialog)
        region_menu.addAction(self.load_regions_action)

        # About (placed in macOS application menu via AboutRole)
        about_action = QAction("About...", self.parent)
        about_action.setMenuRole(QAction.MenuRole.AboutRole)
        about_action.triggered.connect(self.show_about)
        file_menu.addAction(about_action)

    def enable_plane_menu(self, enabled):
        available = bool(enabled)
        self.sub1_action.setEnabled(available)
        self.sub2_action.setEnabled(available)

    def set_undo_redo_enabled(
        self,
        can_undo: bool,
        can_redo: bool,
        *,
        undo_label: str | None = None,
        redo_label: str | None = None,
    ):
        if hasattr(self, "undo_action") and self.undo_action is not None:
            self.undo_action.setEnabled(bool(can_undo))
            text = "Undo Analysis"
            if can_undo and undo_label:
                text = f"Undo Analysis ({undo_label})"
            self.undo_action.setText(text)
        if hasattr(self, "redo_action") and self.redo_action is not None:
            self.redo_action.setEnabled(bool(can_redo))
            text = "Redo Analysis"
            if can_redo and redo_label:
                text = f"Redo Analysis ({redo_label})"
            self.redo_action.setText(text)

    def set_view_navigation_enabled(self, can_back: bool, can_forward: bool):
        if hasattr(self, "view_back_action") and self.view_back_action is not None:
            self.view_back_action.setEnabled(bool(can_back))
        if hasattr(self, "view_forward_action") and self.view_forward_action is not None:
            self.view_forward_action.setEnabled(bool(can_forward))

    def set_wcs_frame_checked(self, frame: str):
        normalized = normalize_display_frame(frame)
        target = self.wcs_frame_actions.get(normalized)
        if target is None:
            target = self.wcs_frame_actions.get(preferred_display_frame(getattr(self.parent, "wcs", None)))
        if target is not None:
            target.setChecked(True)

    def set_wcs_frame(self, frame: str):
        setter = getattr(self.parent, "set_wcs_display_frame", None)
        if callable(setter):
            setter(frame)
        self.set_wcs_frame_checked(frame)

    def set_wcs_decimal_checked(self, use_decimal: bool):
        if use_decimal:
            self.wcs_decimal_action.setChecked(True)
        else:
            self.wcs_sexagesimal_action.setChecked(True)

    def set_wcs_decimal_mode(self, use_decimal: bool):
        setter = getattr(self.parent, "set_wcs_decimal_mode", None)
        if callable(setter):
            setter(bool(use_decimal))
        self.set_wcs_decimal_checked(bool(use_decimal))

    def undo_action_triggered(self):
        if hasattr(self.parent, "undo_last_action"):
            self.parent.undo_last_action()

    def redo_action_triggered(self):
        if hasattr(self.parent, "redo_last_action"):
            self.parent.redo_last_action()

    def view_back_triggered(self):
        if hasattr(self.parent, "view_back"):
            self.parent.view_back()

    def view_forward_triggered(self):
        if hasattr(self.parent, "view_forward"):
            self.parent.view_forward()

    def save_recipe(self):
        if hasattr(self.parent, "save_recipe_dialog"):
            self.parent.save_recipe_dialog()

    def load_recipe(self):
        if hasattr(self.parent, "load_recipe_dialog"):
            self.parent.load_recipe_dialog()

    def save_workspace(self):
        if hasattr(self.parent, "save_workspace"):
            self.parent.save_workspace()
        elif hasattr(self.parent, "save_workspace_dialog"):
            self.parent.save_workspace_dialog()

    def load_workspace(self):
        if hasattr(self.parent, "load_workspace_dialog"):
            self.parent.load_workspace_dialog()

    def open_regrid_panel(self):
        if hasattr(self.parent, "open_regrid_panel"):
            self.parent.open_regrid_panel()
        
    def toggle_control_panel(self):
        if self.control_panel_action.isChecked():
            self.parent.show_control_panel()
        else:
            self.parent.hide_control_panel()
            
    def toggle_range_panel(self):
        if self.range_panel_action.isChecked():
            self.parent.show_range_panel()
        else:
            self.parent.hide_range_panel()


    def toggle_main_window(self):
        if self.main_action.isChecked():
            self.parent.show_main_window()
        else:
            self.parent.hide_main_window()

    def toggle_sub1_window(self):
        if self.sub1_action.isChecked():
            ensure = getattr(self.parent, "ensure_subwindow1", None)
            if callable(ensure):
                sub1 = ensure()
            else:
                sub1 = getattr(self.parent.SubWindow, "subwindow1", None)
            if sub1 is not None:
                sub1.show()
                sub1.raise_()
                sub1.activateWindow()
            else:
                self.sub1_action.setChecked(False)
        else:
            sub1 = getattr(self.parent.SubWindow, "subwindow1", None)
            if sub1 is not None:
                sub1.hide()
    
    def toggle_sub2_window(self):
        if self.sub2_action.isChecked():
            ensure = getattr(self.parent, "ensure_subwindow2", None)
            if callable(ensure):
                sub2 = ensure()
            else:
                sub2 = getattr(self.parent.SubWindow, "subwindow2", None)
            if sub2 is not None:
                sub2.show()
                sub2.raise_()
                sub2.activateWindow()
            else:
                self.sub2_action.setChecked(False)
        else:
            sub2 = getattr(self.parent.SubWindow, "subwindow2", None)
            if sub2 is not None:
                sub2.hide()
        
        
    def _ensure_control_panel(self, *, visible: bool = True):
        ensure = getattr(self.parent, "ensure_control_panel", None)
        if callable(ensure):
            try:
                return ensure(visible=visible)
            except TypeError:
                return ensure()
        return getattr(self.parent, "control_panel", None)

    def _open_control_panel_child(self, panel_attr: str, opener_name: str):
        control_panel = self._ensure_control_panel(visible=False)
        if control_panel is None:
            return
        panel = getattr(control_panel, panel_attr, None)
        if panel is None:
            opener = getattr(control_panel, opener_name, None)
            if callable(opener):
                opener()
            return
        panel.raise_()
        panel.activateWindow()

    def open_colorscale_panel(self):
        self._open_control_panel_child("color_settings_panel", "open_color_settings")
            
    def open_scaling_panel(self):
        self._open_control_panel_child("scaling_panel", "open_scaling_panel")

    def open_unit_conversion_panel(self):
        self._open_control_panel_child("unit_conversion_panel", "open_unit_conversion_panel")
    
    def open_chmap_panel(self):
        self._open_control_panel_child("chmap_settings_panel", "open_chmap_settings")

    def open_clump_finding_panel(self):
        self._open_control_panel_child("clump_finding_panel", "open_clump_finding_panel")

    def open_integ_panel(self):
        self._open_control_panel_child("integ_settings_panel", "open_integ_settings")

    def open_cutout_dialog(self):
        if hasattr(self.parent, 'open_cutout_dialog'):
            self.parent.open_cutout_dialog(use_view_bounds=True)

    def open_mask_panel(self):
        self._open_control_panel_child("mask_settings_panel", "open_mask_settings")

    def open_pvd_panel(self):
        self._open_control_panel_child("pvd_panel", "open_pvd_settings")
    
    def open_spec_window(self):
        self._open_control_panel_child("spec_window", "open_spec_window")

    def open_baseline_panel(self):
        self._open_control_panel_child("baseline_panel", "open_baseline_panel")

    def open_smooth_panel(self):
        self._open_control_panel_child("smooth_settings_panel", "open_smooth_settings")

    def open_contour_panel(self):
        self._open_control_panel_child("contour_panel", "open_contour_panel")


    def open_config_panel(self):
        self.config_panel = ConfigPanel(self.parent.config_manager, self.parent)
        self.config_panel.show()
        
    def open_header_panel(self):
        self.header_panel = ShowHeader(self.parent.header)
        self.header_panel.resize(300, 400)
        self.header_panel.show()
        
    def show_about(self):
        dlg = QMessageBox(self.parent)
        dlg.setWindowTitle("About")
        dlg.setTextFormat(Qt.TextFormat.RichText)
        dlg.setText(
            f'<table cellpadding="4">'
            f'<tr><td><b>{APP_NAME}</b>&nbsp; version {APP_VERSION}</td>'
            f'<td><a href="https://github.com/s-takekawa/takefits">GitHub</a></td></tr>'
            # f'<tr><td>&copy; Shunya Takekawa</td>'
            # f'<td><a href="https://orcid.org/0000-0001-8147-6817">ORCID</a></td></tr>'
            f'</table>'
        )
        dlg.exec()

    def open_arithmetic_panel(self):
        self._open_control_panel_child("arithmetic_panel", "open_arithmetic_panel")


def _copy_menu_contents(source_menu: QMenu, target_menu: QMenu):
    for source_action in list(source_menu.actions() or []):
        if source_action is None:
            continue
        if source_action.isSeparator():
            target_menu.addSeparator()
            continue
        source_submenu = source_action.menu()
        if source_submenu is not None:
            title = str(source_submenu.title() or source_action.text() or "").strip()
            submenu = target_menu.addMenu(title or "Menu")
            submenu.menuAction().setEnabled(source_action.isEnabled())
            submenu.menuAction().setVisible(source_action.isVisible())
            _copy_menu_contents(source_submenu, submenu)
            continue
        target_menu.addAction(source_action)


def mirror_menu_bar_to_window(source_window, target_window) -> bool:
    if source_window is None or target_window is None:
        return False
    if source_window is target_window:
        return False
    source_getter = getattr(source_window, "menuBar", None)
    target_getter = getattr(target_window, "menuBar", None)
    if not callable(source_getter) or not callable(target_getter):
        return False
    try:
        source_bar = source_window.menuBar()
        target_bar = target_window.menuBar()
    except Exception:
        return False
    if source_bar is None or target_bar is None:
        return False
    try:
        target_bar.clear()
    except Exception:
        return False

    for source_action in list(source_bar.actions() or []):
        if source_action is None:
            continue
        source_menu = source_action.menu()
        if source_menu is None:
            target_bar.addAction(source_action)
            continue
        title = str(source_menu.title() or source_action.text() or "").strip()
        mirrored_menu = target_bar.addMenu(title or "Menu")
        mirrored_menu.menuAction().setEnabled(source_action.isEnabled())
        mirrored_menu.menuAction().setVisible(source_action.isVisible())
        _copy_menu_contents(source_menu, mirrored_menu)
    return True



"""Headless visualization usecases."""
from __future__ import annotations

from typing import Optional, Tuple, Union, List

import matplotlib.pyplot as plt
import numpy as np

from takefits.core.app_state import AppState
from takefits.core.usecases.moment import compute_moment
from takefits.core.usecases.channel_map import compute_channel_map, channel_labels_to_world
from takefits.core.usecases.export import export_figure
from takefits.core.usecases.utils import create_2d_header_from_3d
from takefits.core.plotting.display_map import DisplayMap
from takefits.core.config import ConfigManager
from astropy.wcs import WCS
import matplotlib as mpl
from matplotlib import colormaps as mpl_colormaps
from astropy.visualization.wcsaxes import WCSAxes
import astropy.units as u

from takefits.core.custom_colormap import CustomColormap, ColorDefinitions

def register_custom_colormaps():
    """Register custom colormaps (Rainbow, Cool) if not present."""
    # Check if 'Rainbow' is already registered
    if 'Rainbow' not in mpl_colormaps:
        rainbow_cdict = ColorDefinitions.rainbow()
        rainbow = CustomColormap('Rainbow', rainbow_cdict)
        mpl_colormaps.register(rainbow.get_colormap())
        mpl_colormaps.register(rainbow.reversed_colormap().get_colormap())

    if 'Cool' not in mpl_colormaps:
        cool_cdict = ColorDefinitions.cool()
        cool = CustomColormap('Cool', cool_cdict)
        mpl_colormaps.register(cool.get_colormap())
        mpl_colormaps.register(cool.reversed_colormap().get_colormap())


def export_moment_image(
    state: AppState,
    output_path: str,
    moment_type: str = "moment0",
    axis: int = 0,
    pixel_range: Optional[Tuple[float, float]] = None,
    world_range: Optional[Tuple[Union[float, str], Union[float, str]]] = None,
    cmap: str = "viridis",
    origin: str = "lower",
    dpi: int = 150,
    title: Optional[str] = None
) -> str:
    """
    Compute a moment map and export it as an image using CLI/GUI styling.
    
    Args:
        state: AppState with data.
        output_path: Path to save PNG/PDF/etc.
        moment_type: Type of moment to compute (moment0, moment1, etc.)
        axis: Axis to integrate along (0=z).
        pixel_range: Integration range in pixels.
        world_range: Integration range in world coords.
        cmap: Colormap name.
        origin: Image origin ('lower' or 'upper').
        dpi: Output resolution.
        title: Optional title for the plot.
        
    Returns:
        Path to saved file.
    """
    # 1. Compute moment map data
    moment_data = compute_moment(
        state=state,
        moment_type=moment_type,
        axis=axis,
        pixel_range=pixel_range,
        world_range=world_range
    )
    
    # 2. Prepare 2D Header/WCS
    if state.header:
        header_2d = create_2d_header_from_3d(state.header, axis_to_drop=axis)
        try:
            wcs_2d = WCS(header_2d)
        except Exception:
            wcs_2d = None
    else:
        header_2d = None
        wcs_2d = None

    # 3. Configure DisplayMap
    # Load default config
    config_manager = ConfigManager()
    config = config_manager.config.copy()
    
    # Override with user args
    if cmap:
        config['colorscale'] = cmap
    
    # Only force white background if user hasn't customized it? 
    # Actually, for publication plots (headless), user likely *wants* config settings OR specific overrides.
    # We will let the config file dictate styles unless explicitly overridden by CLI args.
    
    # Note: display_map.py uses config keys directly.
    # The 'config' dict now contains everything from config.yaml.
    
    # 4. Render
    # Use config dimension? Or keep fixed size for CLI?
    # GUI config has 'figure_width', 'figure_height', but those are for UI window size.
    # Export usually needs standard size or specific DPI.
    # We stick to fixed figsize for now, but respect Config for fonts/colors/ticks.
    
    fig = plt.figure(figsize=(8, 6))
    
    # Create DisplayMap instance
    # DisplayMap handles WCSAxes, colorbars, ticks, etc.
    dm = DisplayMap(moment_data, header_2d, wcs_2d, config)
    
    # We display on 'xy' plane because we have reduced it to 2D
    dm.display(fig, plane='xy')
    
    if title:
        fig.suptitle(title, y=0.95)

    # 5. Save
    try:
        export_figure(fig, output_path, dpi=dpi)
    finally:
        plt.close(fig)
        
    return output_path


def export_channel_map_image(
    state: AppState,
    output_path: str,
    start_channel: Optional[float] = None,
    end_channel: Optional[float] = None,
    interval: float = 1.0,
    mode: str = "average",
    axis: int = 0,
    ncols: int = 4,
    cmap: Optional[str] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    start_world: Optional[float] = None,
    end_world: Optional[float] = None,
    interval_world: Optional[float] = None,
    dpi: int = 150,
) -> str:
    """
    Export a specific channel map or grid of channel maps to an image file.
    
    Args:
        state: Application state
        output_path: Path to save image
        start_channel: First channel index (0-based)
        end_channel: Last channel index (0-based)
        interval: Step size for channel grid
        mode: 'average' (binning) or 'slice' (step)
        axis: WCS axis index for spectral dimension (default 0)
        ncols: Number of columns in grid
        cmap: Colormap name (default: Rainbow)
        vmin: Minimum intensity value
        vmax: Maximum intensity value
        dpi: Output DPI

    Returns:
        Path to saved file
    """
    register_custom_colormaps()
    
    if cmap is None:
        cmap = 'Rainbow'
    if cmap is None:
        cmap = 'Rainbow'

    # Convert world coordinates to channels if provided
    if start_world is not None and end_world is not None and state.wcs:
        try:
             # Assume spectral axis is axis 2 (0-based index 2 for 3rd axis) for typical cubes
             # We need to construct a world coordinate array.
             # Get CRVALs for spatial axes to fix them
             crval1 = state.header.get('CRVAL1', 0)
             crval2 = state.header.get('CRVAL2', 0)
             
             # Check if CUNIT3 is 'm/s' and input is likely km/s?
             # User standard: "interval 25 km/s code [-150, 150] km/s"
             # If WCS is m/s, we must scale input 1000x.
             cunit3 = state.header.get('CUNIT3', '').lower()
             scale_factor = 1.0
             if 'm/s' in cunit3 and 'km/s' not in cunit3 and abs(start_world) < 10000:
                  # Heuristic: if WCS is m/s but input is small, assume input is km/s
                  scale_factor = 1000.0

             w_start = start_world * scale_factor
             w_end = end_world * scale_factor
             
             # Convert to pixel
             # wcs_world2pix takes (ra, dec, vel) usually
             # We want the pixel index along the spectral axis.
             # We can probe the spectral axis.
             
             # Safe Conversion: Use spectral_coord conversion if possible, or full wcs
             coords_start = [[crval1, crval2, w_start]]
             coords_end = [[crval1, crval2, w_end]]
             
             pix_start = state.wcs.wcs_world2pix(coords_start, 0)[0][2]
             pix_end = state.wcs.wcs_world2pix(coords_end, 0)[0][2]
             
             start_channel = min(pix_start, pix_end)
             end_channel = max(pix_start, pix_end)
             
             if interval_world is not None:
                 # Calculate interval in pixels
                 # Estimate pixel scale at the center velocity?
                 # Or just diff between start and start+interval
                 w_next = (start_world + interval_world) * scale_factor
                 coords_next = [[crval1, crval2, w_next]]
                 pix_next = state.wcs.wcs_world2pix(coords_next, 0)[0][2]
                 interval = abs(pix_next - pix_start)
                 
             # Round to nearest integer for safe slicing/averaging?
             # compute_channel_map handles floats?
             # Usually best to round for clear boundaries unless sub-pixel
             # But 'compute_channel_map' might expect floats for 'slice' mode interpolation?
             # Existing usage implies int/float.
             # Let's leave as float for precision, but compute_channel_map likely casts or uses slicing.
             
        except Exception as e:
            import warnings
            warnings.warn(f"Failed to convert world coordinates: {e}")

    # 1. Compute channel maps
    result = compute_channel_map(
        state=state,
        start_channel=start_channel,
        end_channel=end_channel,
        interval=interval,
        mode=mode,
        axis=axis
    )
    
    images = result.images
    raw_labels = list(getattr(result, "display_labels", result.labels))
    
    # Convert labels to world coordinates if WCS exists
    # If using world coords, we want format like "-150.0"
    # Helper uses axis 0, 1, 2 depending on WCS structure.
    # We need to identify the spectral axis index in the WCS object.
    # 'axis' argument to this function is the DATA axis (usually 0=z).
    # In WCS list, axis order is reversed (x, y, z).
    # So data axis 0 corresponds to WCS axis 2 (usually).
    
    wcs_axis_index = state.data.ndim - 1 - axis if state.data is not None else 2
    
    # However, WCS object might be 2D or 3D.
    # state.wcs usually matches data dimensions.
    
    labels = channel_labels_to_world(state, raw_labels, axis=wcs_axis_index)
    n_images = len(images)
    
    if n_images == 0:
        raise ValueError("No channel maps generated with given parameters.")
        
    # 2. Setup Grid
    # Load config for channel maps (colors, fonts)
    config_manager = ConfigManager()
    config = config_manager.config
    effective_cmap = cmap if cmap else config.get('colorscale', 'viridis')
    nrows = int(np.ceil(n_images / ncols))

    fig = plt.figure(figsize=(ncols * 3, nrows * 3))
    gs = fig.add_gridspec(nrows, ncols)

    
    # Global normalization
    all_data = np.array([img for img in images if img is not None])
    
    if vmin is None:
        vmin = np.nanmin(all_data)
    if vmax is None:
        vmax = np.nanmax(all_data)

    # Prepare WCS projection
    wcs = state.wcs
    slices = None
    if wcs:
        # Determine slices based on axis
        # axis=0 (z) -> xy plane (plane_num=0)
        # axis=1 (y) -> xz plane (plane_num=1)
        # axis=2 (x) -> zy plane (plane_num=2)
        naxis = wcs.naxis
        if naxis == 3:
            if axis == 0: slices = ('x', 'y', 0)
            elif axis == 1: slices = ('x', 0, 'y')
            elif axis == 2: slices = (0, 'y', 'x')
        elif naxis == 4:
            if axis == 0: slices = ('x', 'y', 0, 0)
            elif axis == 1: slices = ('x', 0, 'y', 0)
            elif axis == 2: slices = (0, 'y', 'x', 0)

    # Config values
    tick_color = config.get('tick_color', 'black')
    tick_width = config.get('tick_width', 1)
    tick_labelsize = config.get('tick_labelsize', 10)
    tick_labelcolor = config.get('tick_labelcolor', 'black')
    tick_direction = config.get('tick_direction', 'out')
    tick_length = config.get('tick_length', 4)
    tick_pad_x = config.get('tick_pad_x', 5)
    tick_pad_y = config.get('tick_pad_y', 5)
    
    im_list = []
    
    for i in range(nrows):
        for j in range(ncols):
            idx = i * ncols + j
            
            # WCSAxes if available
            if wcs and slices:
                ax = fig.add_subplot(gs[i, j], projection=wcs, slices=slices)
            else:
                ax = fig.add_subplot(gs[i, j])
                
            if idx < n_images:
                img = images[idx]
                lbl = labels[idx]
                
                # Aspect handling
                if axis == 0: aspect = 'equal' # xy
                else: aspect = 'auto'
                
                im = ax.imshow(img, cmap=effective_cmap, origin='lower', vmin=vmin, vmax=vmax, aspect=aspect)
                im_list.append(im)
                
                # Label
                if len(lbl) == 3:
                    if mode == "integrate":
                        # If string, use as is. If float, format.
                        if isinstance(lbl[1], str): label_text = lbl[1]
                        else: label_text = f"{lbl[1]:.1f}"
                    elif mode == "slice":
                         # Single value
                         if isinstance(lbl[1], (float, int)): label_text = f"{lbl[1]:.1f}"
                         else: label_text = str(lbl[1])
                    else:
                        if isinstance(lbl[1], str): label_text = lbl[1]
                        else: label_text = f"{lbl[1]:.1f}"
                else:
                    label_text = str(lbl)
                
                # Use config font/color for label
                ch_label_font = config.get('ch_label_font', 'Arial')
                ch_label_color = config.get('ch_label_color', 'grey')
                pos_x = config.get('pos_chlabel_x', 0.98)
                pos_y = config.get('pos_chlabel_y', 0.02)
                
                ax.text(pos_x, pos_y, label_text, transform=ax.transAxes, 
                        color=ch_label_color, fontfamily=ch_label_font, 
                        va='bottom', ha='right', fontweight='bold')
                
                # Styling
                ax.tick_params(which='major', direction=tick_direction,
                               length=tick_length, color=tick_color,
                               width=tick_width, labelsize=tick_labelsize,
                               labelcolor=tick_labelcolor)
                
                for spine in ax.spines.values():
                    spine.set_linewidth(tick_width)
                    spine.set_color(tick_color)

                # Consistent tick behavior
                # Logic: Show ticks/labels only on the left edge (col 0) and bottom edge (bottom-most tile in column)
                
                if wcs:
                    # Identify X and Y axes indices in WCS
                    # slices has 'x', 'y' and integers
                    try:
                        x_wcs_axis = slices.index('x')
                        y_wcs_axis = slices.index('y')
                    except ValueError:
                        # Fallback if slices structure is unexpected
                        x_wcs_axis = 0
                        y_wcs_axis = 1

                    is_bottom_tile = (idx + ncols >= n_images)
                    is_left_tile = (j == 0)

                    
                    # Define label mapping (from DisplayMap)
                    coords_dict = {
                        'glon': 'Galactic Longitude',
                        'glat': 'Galactic Latitude',
                        'ra': 'Right Ascension',
                        'dec': 'Declination',
                        # Add frequency/velocity if needed for 3rd axis, though usually not shown on 2D map axes
                    }
                    
                    # Helper to apply format (hms vs deg) based on config/ctype
                    # Similar to DisplayMap.update_axes_format
                    def configure_coord_format(coord, ctype_index):
                        ctype_str = (state.header.get(f'CTYPE{ctype_index+1}', '')).upper() if state.header else ''
                        # Basic formatting logic
                        is_decimal = config.get('decimal', False)
                        
                        if ctype_str.startswith('RA'):
                            wrap_val = config.get('coord_wrap', 360.0)
                            if isinstance(wrap_val, (int, float)):
                                coord.set_coord_type('longitude', coord_wrap=float(wrap_val) * u.deg)
                            else:
                                coord.set_coord_type('longitude', coord_wrap=360.0 * u.deg)
                            
                            if is_decimal:
                                coord.set_format_unit(u.deg, decimal=True)
                            else:
                                coord.set_format_unit(u.hourangle, decimal=False)
                                
                        elif ctype_str.startswith('DEC'):
                             coord.set_coord_type('latitude')
                             if is_decimal:
                                 coord.set_format_unit(u.deg, decimal=True)
                                 coord.set_format_unit(u.deg, decimal=False)
                        
                        elif any(keyword in ctype_str for keyword in ['GLON', 'GLAT', 'OFFSET']):
                            if 'GLON' in ctype_str:
                                wrap_val = config.get('coord_wrap', 180)
                                if isinstance(wrap_val, (int, float)):
                                     coord.set_coord_type('longitude', coord_wrap=float(wrap_val) * u.deg)
                                else:
                                     coord.set_coord_type('longitude', coord_wrap=180 * u.deg)
                            else: 
                                coord.set_coord_type('latitude')
                            
                            coord.set_format_unit('deg', decimal=is_decimal)

                    for k in range(len(slices)):
                        configure_coord_format(ax.coords[k], k)

                    # Apply Labels Globally to the axes object first (so the text is correct)
                    for alias, label in coords_dict.items():
                        if alias in ax.coords:
                            ax.coords[alias].set_axislabel(label, 
                                fontsize=config.get('axislabel_fontsize', 14),
                                fontfamily=config.get('axislabel_fontfamily', 'Arial'),
                                color=config.get('axislabel_color', 'black'))

                    # Apply Visibility
                    for k in range(len(slices)):
                        coord = ax.coords[k]
                        
                        # Handle X Axis (Horizontal)
                        if k == x_wcs_axis:
                            if is_bottom_tile:
                                coord.set_ticklabel_visible(True)
                                coord.set_ticks_position(config.get('xticklabel_position', 'b'))
                                coord.set_ticklabel_position(config.get('xticklabel_position', 'b'))
                            else:
                                coord.set_axislabel('') # Hide label text on internal
                                coord.set_ticklabel_visible(False)
                        
                        # Handle Y Axis (Vertical)
                        elif k == y_wcs_axis:
                            if is_left_tile:
                                coord.set_ticklabel_visible(True)
                                coord.set_ticks_position(config.get('yticklabel_position', 'l'))
                                coord.set_ticklabel_position(config.get('yticklabel_position', 'l'))
                            else:
                                coord.set_axislabel('') # Hide label text on internal
                                coord.set_ticklabel_visible(False)
                        
                        # Handle Sliced Axes (should be hidden)
                        else:
                             coord.set_axislabel('')
                             coord.set_ticklabel_visible(False)
                             coord.set_ticks_visible(False)

                    # Ensure ticks themselves are visible everywhere

                    # Ensure ticks themselves are visible everywhere (often desired) or just edges?
                    # GUI "default_ticks_position" usually means ticks on all sides?
                    # If config says 'btlr', we put ticks everywhere.
                    ticks_pos = config.get('default_ticks_position', 'btlr')
                    for coord in ax.coords:
                        coord.set_ticks_position(ticks_pos)
                        coord.display_minor_ticks(True)

            else:
                ax.axis('off')
    
    # Global Axis Labels (Fake it by placing text on the figure edge?)
    # Or rely on WCSAxes labels on the edge subplots.
    # WCSAxes automatically adds labels if set_axislabel is not empty.
    # We cleared them above. Let's restore them just for the "center" of the edges?
    # Or just let every edge plot have the label? 
    # Usually "Declination" appears on the Y axis of the left column.
    
    # Let's simple enable axis labels on the edges where we enabled tick labels.
    # But this might be redundant if every row has "Declination".
    # Standard publication style: Label on the center-left and center-bottom.
    
    # We will let WCSAxes show labels on all edge plots for now, as that's robust.
    # Re-loop to set axis labels on edges?
    # Or just do it in the loop above.
    
    # Refined loop logic for edges:
    # We need to access ax.coords again.
    
    # Let's rely on standard WCS defaults but just hide them internally.
    # (The loop above mostly did this, but let's be explicit about text labels)
    
    # Colorbar
    # Create separate axes for colorbar like GUI
    cbar_pos_x = config.get('cbar_pos_x', 0.9)
    cbar_pos_y = config.get('cbar_pos_y', 0.11)
    cbar_width = config.get('cbar_width', 0.04)
    cbar_height = config.get('cbar_height', 0.77)
    
    # Note: These GUI positions are relative to window/figure. 
    # With tight_layout/constrained_layout, fixed axes might overlap or look weird.
    # However, user requested GUI style matching.
    # The GUI uses fixed axes positions for the colorbar on the figure.
    # If we do fig.add_axes([...]) it is absolute figure coordinates (0-1).
    
    # We used subplots inside GridSpec above.
    # To match GUI exactly, we might need to be careful.
    
    # Let's try adding it to the right of the grid using standard matplotlib first?
    # NO, user wants "SAME STYLE". GUI has it at fixed position on right.
    
    # But wait, GUI layout logic:
    # layout.addWidget(self.canvas, 2, 0, 1, 16)
    # The canvas is the widget. The figure is inside.
    # The config coords (0.9, 0.11) seem to be Figure coordinates.
    
    if im_list:
        cbar_orientation = config.get('colorbar_orientation', 'vertical')
        cax = fig.add_axes([cbar_pos_x, cbar_pos_y, cbar_width, cbar_height])
        
        cbar = fig.colorbar(im_list[0], cax=cax, orientation=cbar_orientation)
        
        # Colorbar Ticks
        cax.tick_params(
             axis='y', which='both',
             left=config.get('colorbar_tick_left', False),
             right=config.get('colorbar_tick_right', True),
             labelleft=config.get('colorbar_tick_labelleft', False),
             labelright=not config.get('colorbar_tick_labelleft', False),
             width=config.get('colorbar_tick_width', 1),
             length=config.get('colorbar_tick_length', 2),
             color=config.get('colorbar_tick_color', 'black'),
             direction=config.get('colorbar_tick_direction', 'out'),
             labelcolor=config.get('colorbar_tick_labelcolor', 'black')
        )
        # (Add x axis params if needed, omitted for brevity as vertical is standard)
        
        cbar.outline.set_color(config.get('colorbar_tick_color', 'black'))
        cbar.outline.set_linewidth(config.get('colorbar_tick_width', 1))

    if im_list:
        right_margin = cbar_pos_x - 0.02 if cbar_pos_x < 1.0 else 0.88
    else:
        right_margin = 0.95
    right_margin = float(np.clip(right_margin, 0.55, 0.98))
    fig.subplots_adjust(left=0.08, right=right_margin, bottom=0.08, top=0.95, wspace=0.12, hspace=0.12)

    # 3. Save
    try:
        export_figure(fig, output_path, dpi=dpi)
    finally:
        plt.close(fig)
        
    return output_path



import matplotlib as mpl
import numpy as np

class AddHPBW:
    initial_config = None
    def __init__(self, ax, header, config):
        self.ax = ax
        self.header = header
        if AddHPBW.initial_config is None:
            AddHPBW.initial_config = config
        self.config = AddHPBW.initial_config
        self.ellipse = None
        
        self.create_beam()

        

    def create_beam(self):
        hpbw_major_pix, hpbw_minor_pix, bpa = self.calculate_hpbw_in_pixels()
        if self.ellipse:
            self.ellipse.remove()
        self.ellipse = mpl.patches.Ellipse(
            (0, 0),
            width=hpbw_major_pix, height=hpbw_minor_pix,
            angle=bpa+90., edgecolor=self.config.get('beam_edgecolor', 'None'), facecolor=self.config.get('beam_facecolor', 'white'), lw=self.config.get('beam_linewidth', 0)
        )
        self.ax.add_patch(self.ellipse)
        self.update_position()

        self.ax.figure.canvas.mpl_connect("draw_event", self.update_position)
        self.ax.figure.canvas.mpl_connect("motion_notify_event", self.update_position)
        # Wrap savefig to temporarily show HPBW ellipse when saving
        fig = self.ax.figure
        orig_savefig = fig.savefig
        def savefig_with_beam(*args, **kwargs):
            self.ellipse.set_visible(True)
            try:
                return orig_savefig(*args, **kwargs)
            finally:
                self.ellipse.set_visible(False)
        fig.savefig = savefig_with_beam

    def refresh_geometry_from_header(self):
        """Refresh beam size/angle from current FITS header without recreating hooks."""
        if self.ellipse is None:
            self.create_beam()
            return
        hpbw_major_pix, hpbw_minor_pix, bpa = self.calculate_hpbw_in_pixels()
        self.ellipse.width = hpbw_major_pix
        self.ellipse.height = hpbw_minor_pix
        self.ellipse.angle = bpa + 90.0
        self.update_position()


    def calculate_hpbw_in_pixels(self):
        cunit1 = self.header.get('CUNIT1', '').strip().lower()
        cunit2 = self.header.get('CUNIT2', '').strip().lower()
        if cunit1 == '' and cunit2 == '': pass
        elif cunit1 != 'deg' or cunit2 != 'deg':
            #raise ValueError("CUNIT1 and CUNIT2 must be in degrees (deg).")
            return 0, 0, 0
    
        bmaj = self.header.get('BMAJ', 0)
        bmin = self.header.get('BMIN', 0)
        bpa = self.header.get('BPA', 0)
        
        try:
            cdelt1 = abs(self.header['CDELT1'])
            cdelt2 = abs(self.header['CDELT2'])
        except: #to be developed!
            print("\033[1;31m\033[1mWarning: Coordinates are defined by CD matrix.\033[0m")
            print("\033[1;31m\033[1mCoordinate axis is tilted to the frame.\033[0m")
            cdelt1 = np.sqrt((self.header['CD1_1'])**2 + (self.header['CD2_1'])**2)
            cdelt2 = np.sqrt((self.header['CD1_2'])**2 + (self.header['CD2_2'])**2)

        #if bmaj is None or bmin is None:
        #    raise ValueError("BMAJ or BMIN is not available in the FITS header.")

        hpbw_major_pix = bmaj / cdelt1
        hpbw_minor_pix = bmin / cdelt2

        return hpbw_major_pix, hpbw_minor_pix, bpa

    def update_position(self, event=None):
        self.relative_x = self.config.get('beam_pos_x', 0.1)
        self.relative_y = self.config.get('beam_pos_y', 0.1)
        xlim = self.ax.get_xlim()
        ylim = self.ax.get_ylim()
        center_x = xlim[0] + self.relative_x * (xlim[1] - xlim[0])
        center_y = ylim[0] + self.relative_y * (ylim[1] - ylim[0])
        self.ellipse.set_center((center_x, center_y))
        if not self.ellipse._visible: self.ellipse.set_visible(True)
        self.ellipse.set_facecolor(self.config.get('beam_facecolor'))
        self.ellipse.set_edgecolor(self.config.get('beam_edgecolor'))
        self.ellipse.set_linewidth(self.config.get('beam_linewidth'))
        try:
            self.ax.draw_artist(self.ellipse)
        except AttributeError:
            # skip backends without renderer (e.g. PDF)
            pass
        self.ellipse.set_visible(False)
        
    def update_ax(self, new_ax):
        self.ax = new_ax
        self.create_beam()



from PySide6.QtWidgets import QWidget, QVBoxLayout, QGridLayout, QComboBox, QLineEdit, QPushButton, QCheckBox, QLabel, QToolButton, QSizePolicy, QMessageBox, QDoubleSpinBox
from PySide6.QtGui import QPalette
from PySide6.QtCore import Qt, QSignalBlocker, QTimer, QSize
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import colormaps
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas

from takefits.core.color import ColorMode, default_color_settings_map
from takefits.core.custom_colormap import CustomColormap, ColorDefinitions
from takefits.logic.data_tools import (
    estimate_array_nbytes,
    fast_nanminmax,
    MEMMAP_THRESHOLD_BYTES,
)


class RegisterColor:
    rainbow_cdict = ColorDefinitions.rainbow()
    cool_cdict = ColorDefinitions.cool()
    rainbow = CustomColormap('Rainbow', rainbow_cdict)
    cool = CustomColormap('Cool', cool_cdict)
    rainbow_r = rainbow.reversed_colormap()
    cool_r = cool.reversed_colormap()
    colormaps.register(rainbow.get_colormap())
    colormaps.register(cool.get_colormap())
    colormaps.register(rainbow_r.get_colormap())
    colormaps.register(cool_r.get_colormap())



class ColorSettingsPanel(QWidget):
    bad_color = None
    filename = None
    
    settings = default_color_settings_map()

    def __init__(self, fits_viewer, subwindows=None, data=None, config=None, color_pattern=None, bad_color=None, filename = None, mode = ColorMode.MAIN):
        super().__init__()
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.fits_viewer = fits_viewer
        self.subwindows = subwindows
        self.mode = mode
        self._histogram_artist = None
        self._shared_history_record_pending = False
        self._shared_history_record_reason = ""
        self._suppress_shared_history_recording = True
        self._pending_overlay_refresh_cids = {}
        self.current_settings = ColorSettingsPanel.settings[self.mode]
        
        if config is None: self.config = self.fits_viewer.displaymap.config
        else: self.config = config
        self.dragging_min = False
        self.dragging_max = False
        if data is None: self.data = self.fits_viewer.data
        else: self.data = data
        self._data_nbytes = estimate_array_nbytes(self.data)
        if  color_pattern is None: self.color_pattern = self.fits_viewer.displaymap.colorscale
        elif self.current_settings["color_pattern"] is None: self.color_pattern = color_pattern
        if self.current_settings["color_pattern"] is not None:
            self.color_pattern = self.current_settings["color_pattern"]

        if bad_color is None: 
            self.bad_color = self.fits_viewer.displaymap.bad_color
        else: self.bad_color = bad_color
        if filename is None:
            self.filename = self.fits_viewer.filename
        else: self.filename = filename
        
        if mode == ColorMode.MAIN:
            self.hist_color = 'gold'
            self.vline_color = 'orange'
        elif mode == ColorMode.INTEG:
            self.hist_color = 'skyblue'
            self.vline_color = 'deepskyblue'
        elif mode == ColorMode.PV:
            self.hist_color = 'mediumturquoise'
            self.vline_color = 'teal'
        elif mode == ColorMode.CHANNEL:
            self.hist_color = 'yellowgreen'
            self.vline_color = 'olivedrab'
        self.initUI()


    def initUI(self):
        self.layout = QVBoxLayout()
        palette = self.palette()
        background_color = palette.color(QPalette.ColorRole.Window).name()

        self.fig, self.ax = plt.subplots(figsize=(4., 2.), facecolor=background_color)
        self.canvas = FigureCanvas(self.fig)
        self.tick_color = 'white' if palette.color(QPalette.ColorRole.Window).lightness() < 128 else 'black'
        self.ax.tick_params(axis='both', which='major', labelsize=8,  colors=self.tick_color)
        self.ax.tick_params(axis='both', which='minor', labelsize=8, colors=self.tick_color) 
        for spine in self.ax.spines.values():
            spine.set_edgecolor(self.tick_color)
        self.ax.set_xlabel('Intensity', fontsize = 8, color=self.tick_color)
        self.ax.set_ylabel('Count', fontsize = 8, color=self.tick_color)
        
        
        # Controls
        self.controls_layout = QGridLayout()
        self.controls_layout.setHorizontalSpacing(6)
        self.controls_layout.setVerticalSpacing(3)
        self.controls_layout.setContentsMargins(0, 0, 0, 0)
        
        self.colorscale_combo = QComboBox(self)
        self.colorscale_combo.setEditable(True)
        colormap_names = [name for name in colormaps.keys() if not name.endswith('_r')]
        self.colorscale_combo.addItems(sorted(colormap_names))
        
        self.colorscale_combo.currentIndexChanged.connect(self.change_color_scale)
        self.colorscale_combo.setFixedWidth(100)
        self.controls_layout.addWidget(self.colorscale_combo, 0, 0)
        
        self.intensity_min = QLineEdit(self)
        self.intensity_min.setPlaceholderText('min')
        self.intensity_min.returnPressed.connect(self.update_intensity_range)
        self.controls_layout.addWidget(self.intensity_min, 0, 1, 1, 2)
        
        self.intensity_max = QLineEdit(self)
        self.intensity_max.setPlaceholderText('max')
        self.intensity_max.returnPressed.connect(self.update_intensity_range)
        self.controls_layout.addWidget(self.intensity_max, 0, 3, 1, 2)

        if self.current_settings["min_val"] is None or self.current_settings["max_val"] is None:
            display_min, display_max = self._current_display_range()
            if display_min is not None and display_max is not None:
                self.current_settings["min_val"] = display_min
                self.current_settings["max_val"] = display_max

        if self.current_settings["min_val"] is not None:
            self.intensity_min.setText(f"{self.current_settings['min_val']:.3g}")
        if self.current_settings["max_val"] is not None:
            self.intensity_max.setText(f"{self.current_settings['max_val']:.3g}")
        
        self.set_button = QPushButton('Set', self)
        self.set_button.clicked.connect(self.update_intensity_range)
        self.controls_layout.addWidget(self.set_button, 2, 3, 1, 2)
        
        self.min_max_button = QPushButton('Min・Max', self)
        self.min_max_button.clicked.connect(self.set_min_max)
        self.controls_layout.addWidget(self.min_max_button, 1, 1, 1, 2)
        
        self.auto_button = QPushButton('Auto', self)
        self.auto_button.clicked.connect(self.auto_intensity)
        self.controls_layout.addWidget(self.auto_button, 1, 3, 1, 2)  
        

        
        self.invert_checkbox = QCheckBox('Invert', self)
        self.invert_checkbox.stateChanged.connect(self.change_color_scale)
        self.controls_layout.addWidget(self.invert_checkbox, 1, 0)
        
        self.log_checkbox = QCheckBox('Log', self)
        self.log_checkbox.clicked.connect(self.toggle_log_scale)
        
        
        self.controls_layout.addWidget(self.log_checkbox, 2, 0, 1, 2)
        
        
        
        self.gamma_label = QLabel('gamma:', self)
        self.controls_layout.addWidget(self.gamma_label, 2, 1, 1, 1, alignment=Qt.AlignmentFlag.AlignRight)
        self.gamma_spinbox = QDoubleSpinBox(self)
        self.gamma_spinbox.setRange(0.0, 100.0)
        self.gamma_spinbox.setFixedWidth(50)
        self.gamma_spinbox.setSingleStep(0.1)
        self.gamma_spinbox.setDecimals(1)
        self.gamma_spinbox.setValue(self.current_settings["gamma_value"])
        
        self.controls_layout.addWidget(self.gamma_spinbox, 2, 2, 1, 1, alignment=Qt.AlignmentFlag.AlignLeft)
        self.gamma_spinbox.editingFinished.connect(self.update_gamma_from_spinbox)
        self.set_button.clicked.connect(self.update_gamma_from_spinbox)
        
        self.hist_toggle_button = QToolButton(self)
        self.hist_toggle_button.setCheckable(True)
        self.hist_toggle_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.hist_toggle_button.setArrowType(Qt.ArrowType.RightArrow)
        self.hist_toggle_button.setIconSize(QSize(8, 8))
        self.hist_toggle_button.setText(" Histogram")
        self.hist_toggle_button.clicked.connect(self.on_toggle)
        self.hist_toggle_button.setStyleSheet("QToolButton { border: none; padding: 0px; font-size: 13px;}")

        
        self.content_area = QWidget(self)
        self.content_area.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        

        self.content_area_layout = QVBoxLayout()
        self.content_area_layout.setContentsMargins(0, 0, 0, 0)
        self.content_area.setLayout(self.content_area_layout)
        self.content_area_layout.addWidget(self.canvas)

        self.layout.addWidget(self.hist_toggle_button)
        self.layout.addWidget(self.content_area)
        self.layout.addLayout(self.controls_layout)

        self.layout.setContentsMargins(10, 10, 10, 0)

        self.setLayout(self.layout)
        self.setWindowTitle(f'Color Settings:{self.filename}')
        
        allow_auto_histogram = (
            self.data.size < 5e8
            and (
                self._data_nbytes is None
                or self._data_nbytes <= MEMMAP_THRESHOLD_BYTES
            )
        )
        if allow_auto_histogram:
            self.hist_toggle_button.setChecked(True)
            self.hist_toggle_button.setArrowType(Qt.ArrowType.DownArrow)
            self.content_area.setVisible(True)
            QTimer.singleShot(0, self.on_toggle)
        else:
            self.hist_toggle_button.setChecked(False)
            self.content_area.setVisible(False)
        
        self.adjustSize()
        
        
        # Initial histogram update
        blockers = [
            QSignalBlocker(self.colorscale_combo),
            QSignalBlocker(self.invert_checkbox),
            QSignalBlocker(self.log_checkbox),
            QSignalBlocker(self.gamma_spinbox),
        ]
        try:
            self.colorscale_combo.setCurrentText(self.color_pattern)
            self.init_vertical_lines()
            self.invert_checkbox.setChecked(self.current_settings["invert"])
            self.log_checkbox.setChecked(self.current_settings["log_scale"])
            self.gamma_spinbox.setValue(self.current_settings["gamma_value"])
        finally:
            del blockers

        # Connect canvas events for dragging
        self.canvas.mpl_connect('button_press_event', self.on_click)
        self.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.canvas.mpl_connect('motion_notify_event', self.on_drag)
        self.canvas.mpl_connect('button_release_event', self.on_release)
        self._suppress_shared_history_recording = False


        
    def on_toggle(self):
        checked = self.hist_toggle_button.isChecked()
        if checked:
            self.hist_toggle_button.setArrowType(Qt.ArrowType.DownArrow)
            self.update_histogram()
            self.content_area.setVisible(True)
        else:
            self.hist_toggle_button.setArrowType(Qt.ArrowType.RightArrow)
            self.content_area.setVisible(False)

    def _resolve_main_viewer(self):
        viewer = getattr(self, "fits_viewer", None)
        if viewer is None:
            return None
        main_viewer = getattr(viewer, "main_viewer", None)
        if main_viewer is not None:
            return main_viewer
        root = getattr(viewer, "fits_viewer", None)
        if root is None:
            return viewer
        nested = getattr(root, "main_viewer", None)
        return nested if nested is not None else root

    def _iter_history_record_targets(self):
        root_viewer = self._resolve_main_viewer()
        targets = []
        if self.mode == ColorMode.MAIN:
            if root_viewer is not None:
                targets.append(root_viewer)
            return targets

        if self.mode == ColorMode.INTEG and root_viewer is not None:
            for ref in list(getattr(root_viewer, "integ_result_windows", []) or []):
                window = ref() if callable(ref) else ref
                if window is None:
                    continue
                targets.append(window)
        elif self.mode == ColorMode.CHANNEL and root_viewer is not None:
            for window in list(getattr(root_viewer, "channel_map_windows", []) or []):
                if window is None:
                    continue
                targets.append(window)

        viewer = getattr(self, "fits_viewer", None)
        if viewer is not None:
            targets.insert(0, viewer)

        deduped = []
        seen = set()
        for target in targets:
            marker = id(target)
            if marker in seen:
                continue
            seen.add(marker)
            deduped.append(target)
        return deduped

    def _schedule_shared_color_history(self, reason: str = "color"):
        if bool(getattr(self, "_suppress_shared_history_recording", False)):
            return
        self._shared_history_record_reason = str(reason or "color")
        if self._shared_history_record_pending:
            return
        self._shared_history_record_pending = True
        QTimer.singleShot(0, self._flush_shared_color_history_record)

    def _flush_shared_color_history_record(self):
        self._shared_history_record_pending = False
        if bool(getattr(self, "_suppress_shared_history_recording", False)):
            return
        reason = f"color:{self._shared_history_record_reason or 'color'}"
        targets = self._iter_history_record_targets()
        if not targets:
            return

        for target in targets:
            if self.mode == ColorMode.MAIN:
                recorder = getattr(target, "_record_shared_view_history", None)
            else:
                recorder = getattr(target, "_record_local_view_history", None)
            if not callable(recorder):
                continue
            try:
                recorder(reason=reason)
            except Exception:
                continue
    
    def move_to_default_position(self):
        control_panel_geometry = self.fits_viewer.control_panel.geometry()
        control_panel_x = control_panel_geometry.x()
        control_panel_y = control_panel_geometry.y()
        control_panel_width = control_panel_geometry.width()

        # Move ColorSettingsPanel to the right of ControlPanel
        self.move(control_panel_x + control_panel_width, control_panel_y)

    def _current_display_range(self):
        try:
            if self.mode == ColorMode.CHANNEL:
                im = self.fits_viewer.im_list[0] if getattr(self.fits_viewer, 'im_list', None) else None
            else:
                im = getattr(self.fits_viewer, 'im', None)
            if im is None:
                return (None, None)

            clim = im.get_clim()
            if clim is None:
                return (None, None)

            vmin, vmax = clim
            if vmin is None or vmax is None:
                return (None, None)

            if not (np.isfinite(vmin) and np.isfinite(vmax)):
                return (None, None)

            return (float(vmin), float(vmax))
        except Exception:
            return (None, None)

    def _is_large_data_mode(self):
        viewers = [
            getattr(self, "fits_viewer", None),
            getattr(getattr(self, "fits_viewer", None), "main_viewer", None),
            getattr(getattr(self, "fits_viewer", None), "fits_viewer", None),
        ]
        seen = set()
        for viewer in viewers:
            if viewer is None:
                continue
            marker = id(viewer)
            if marker in seen:
                continue
            seen.add(marker)
            checker = getattr(viewer, "is_large_data_mode", None)
            if not callable(checker):
                continue
            try:
                return bool(checker())
            except Exception:
                continue
        return False

    def _current_image_data_range(self):
        try:
            if self.mode == ColorMode.CHANNEL:
                im = self.fits_viewer.im_list[0] if getattr(self.fits_viewer, 'im_list', None) else None
            else:
                im = getattr(self.fits_viewer, 'im', None)
            if im is None:
                return (None, None)

            data = im.get_array()
            if data is None:
                return (None, None)
            if np.ma.isMaskedArray(data):
                data = data.filled(np.nan)
            data = np.asarray(data)
            if data.size == 0:
                return (None, None)

            with np.errstate(all='ignore'):
                finite = data[np.isfinite(data)]
                if finite.size == 0:
                    return (None, None)
                data_min = float(np.min(finite))
                data_max = float(np.max(finite))

            if not np.isfinite(data_min) or not np.isfinite(data_max):
                return (None, None)
            if data_max < data_min:
                data_max = data_min
            return (data_min, data_max)
        except Exception:
            return (None, None)

    def _data_range(self):
        if self._is_large_data_mode():
            display_data_min, display_data_max = self._current_image_data_range()
            if display_data_min is not None and display_data_max is not None:
                return (display_data_min, display_data_max)

            display_min, display_max = self._current_display_range()
            if display_min is not None and display_max is not None:
                return (display_min, display_max)

        allow_full_scan = (
            self._data_nbytes is None
            or self._data_nbytes <= MEMMAP_THRESHOLD_BYTES
        )
        if allow_full_scan:
            try:
                with np.errstate(all='ignore'):
                    data_min = float(np.nanmin(self.data))
                    data_max = float(np.nanmax(self.data))
            except ValueError:
                data_min = 0.0
                data_max = 0.0
        else:
            data_min, data_max = fast_nanminmax(self.data)

        if not np.isfinite(data_min):
            data_min = 0.0

        if not np.isfinite(data_max):
            data_max = data_min

        if data_max < data_min:
            data_max = data_min

        return (data_min, data_max)

    def update_intensity_range(self):
        if self.mode == ColorMode.CHANNEL:
            if self.fits_viewer.im_list[0].get_cmap().name != 'from_list':
                self.change_color_scale()
        else:
            if self.fits_viewer.im.get_cmap().name != 'from_list':
                self.change_color_scale()
        try:
            data_min, data_max = self._data_range()
            min_val = float(self.intensity_min.text()) if self.intensity_min.text() else data_min
            max_val = float(self.intensity_max.text()) if self.intensity_max.text() else data_max
            if min_val > max_val:
                min_val, max_val = max_val, min_val
                self.intensity_min.setText(str(f"{min_val:.3g}"))
                self.intensity_max.setText(str(f"{max_val:.3g}"))
                    
            self.current_settings["min_val"] = min_val
            self.current_settings["max_val"] = max_val
            
            self.apply_to_all_windows(lambda im: im.set_clim(min_val, max_val))
            
            # Update histogram lines
            self.update_histogram_lines(min_val, max_val)
            self._schedule_shared_color_history("intensity")
            
            """
            # Save the state
            if not self.integ:
                ColorSettingsPanel.min_val = min_val
                ColorSettingsPanel.max_val = max_val
            else:
                ColorSettingsPanel.min_val_integ = min_val
                ColorSettingsPanel.max_val_integ = max_val
            """
            
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Invalid intensity range values')
    
    def set_min_max(self):
        min_val, max_val = self._data_range()
        self.intensity_min.setText(str(f"{min_val:.3g}"))
        self.intensity_max.setText(str(f"{max_val:.3g}"))
        self.update_intensity_range()

    def auto_intensity(self):
        data_min, data_max = self._data_range()
        max_val = data_max * 0.8 if np.isfinite(data_max) else data_max
        min_val = max(data_min, 0.0) if np.isfinite(data_min) else data_min

        #max_val = np.nanpercentile(self.data[self.data > 0], 99.99)
        self.intensity_min.setText(str(f"{min_val:.3g}"))
        self.intensity_max.setText(str(f"{max_val:.3g}"))
        self.update_intensity_range()
    
    def toggle_log_scale(self):
        try:
            min_val = float(self.intensity_min.text())
            max_val = float(self.intensity_max.text())
    
            if self.log_checkbox.isChecked():
                if min_val <= 0 or max_val <= 0:
                    raise ValueError("Intensity range values must be positive for log scale.")
                
                self.apply_to_all_windows(lambda im: im.set_norm(mpl.colors.LogNorm(vmin=min_val, vmax=max_val)))
                self.apply_to_all_windows(lambda im: im.cmap.set_bad(color=self.config.get('bad_color', 'black')))
                self.auto_button.setEnabled(False)
                self.min_max_button.setEnabled(False)
                #ColorSettingsPanel.log_scale = True
            else:
                norm = mpl.colors.Normalize(vmin=min_val, vmax=max_val)
                self.apply_to_all_windows(lambda im: im.set_norm(norm))
                self.auto_button.setEnabled(True)
                self.min_max_button.setEnabled(True)
                #ColorSettingsPanel.log_scale = False    
        except ValueError as e:
            QMessageBox.warning(self, 'Error', str(e))
            self.log_checkbox.setChecked(False)
            
            #ColorSettingsPanel.log_scale = False
        self.current_settings["log_scale"] = self.log_checkbox.isChecked()
        self.update_intensity_range()
    
    def apply_to_all_windows(self, func):        
        self._suppress_colorbar_layout_for_pending_update()
        try:
            if self.mode == ColorMode.CHANNEL:
                for i in range(len(self.fits_viewer.im_list)):
                    im = self.fits_viewer.im_list[i]
                    func(im)
                    self._invalidate_viewer_cache(self.fits_viewer)
                    if getattr(self.fits_viewer, 'canvas', None):
                        self.fits_viewer.canvas.draw_idle()
                    if self.subwindows is not None:
                        for window in self.subwindows:
                            if window:
                                func(window.im_list[i])       
                                self._invalidate_viewer_cache(window)
                                if getattr(window, 'canvas', None):
                                    window.canvas.draw_idle()
            else:
                func(self.fits_viewer.im)
                self._invalidate_viewer_cache(self.fits_viewer)
                if getattr(self.fits_viewer, 'canvas', None):
                    self.fits_viewer.canvas.draw_idle()
                if self.subwindows is not None:
                    for window in self.subwindows:
                        if window:
                            func(window.im)       
                            self._invalidate_viewer_cache(window)
                            if getattr(window, 'canvas', None):
                                window.canvas.draw_idle()

            self._reapply_colorbar_style_for_changed_viewers(self._current_colorbar_config())
            self._refresh_overlay_after_color_change()
            self._schedule_overlay_refresh_after_next_draw()
        finally:
            self._defer_restore_colorbar_layout_after_pending_update()

    def _invalidate_viewer_cache(self, viewer):
        if viewer is None:
            return
        state = getattr(viewer, 'state', None)
        if state is not None:
            state.image_background = None
            state._background = None
        try:
            viewer._background = None
        except Exception:
            pass

    def _suppress_colorbar_layout_for_pending_update(self):
        if getattr(self, "_pending_layout_suppression", None) is not None:
            return
        targets = []
        viewer = getattr(self, "fits_viewer", None)
        if viewer is not None:
            targets.append(viewer)
        for window in list(getattr(self, "subwindows", None) or []):
            if window is not None:
                targets.append(window)

        records = []
        seen = set()
        for target in targets:
            marker = id(target)
            if marker in seen:
                continue
            seen.add(marker)
            previous = self._set_colorbar_layout_suppressed(target, True)
            if previous is not None:
                records.append((target, previous))
        self._pending_layout_suppression = records

    def _restore_colorbar_layout_after_pending_update(self):
        records = list(getattr(self, "_pending_layout_suppression", None) or [])
        self._pending_layout_suppression = None
        for target, previous in records:
            try:
                setattr(target, "_suspend_colorbar_auto_layout", bool(previous))
            except Exception:
                continue

    def _iter_changed_viewers_for_color_update(self):
        viewers = []
        root = getattr(self, "fits_viewer", None)
        if root is not None:
            viewers.append(root)
        for window in list(getattr(self, "subwindows", None) or []):
            if window is not None:
                viewers.append(window)

        deduped = []
        seen = set()
        for viewer in viewers:
            marker = id(viewer)
            if marker in seen:
                continue
            seen.add(marker)
            deduped.append(viewer)
        return deduped

    def _redraw_viewer_overlay_after_color_change(self, viewer):
        if viewer is None:
            return
        redraw_for_plane = getattr(viewer, "redraw_overlay_for_plane", None)
        if callable(redraw_for_plane):
            plane = str(getattr(viewer, "plane", "") or "")
            try:
                if plane in ("xy", "xz", "zy"):
                    redraw_for_plane(plane)
                else:
                    redraw_for_plane()
                return
            except Exception:
                pass
        redraw_main = getattr(viewer, "redraw_main_overlay_and_blit", None)
        if callable(redraw_main):
            try:
                redraw_main()
                return
            except Exception:
                pass
        canvas = getattr(viewer, "canvas", None)
        if canvas is not None:
            try:
                canvas.draw_idle()
            except Exception:
                pass

    def _schedule_overlay_refresh_after_next_draw(self):
        for viewer in self._iter_changed_viewers_for_color_update():
            canvas = getattr(viewer, "canvas", None)
            if canvas is None:
                continue
            marker = id(canvas)
            if marker in self._pending_overlay_refresh_cids:
                continue

            def _on_draw(_event, *, marker=marker, canvas=canvas, viewer=viewer):
                cid = self._pending_overlay_refresh_cids.pop(marker, None)
                if cid is not None:
                    try:
                        canvas.mpl_disconnect(cid)
                    except Exception:
                        pass
                self._redraw_viewer_overlay_after_color_change(viewer)

            try:
                cid = canvas.mpl_connect("draw_event", _on_draw)
            except Exception:
                continue
            self._pending_overlay_refresh_cids[marker] = cid

    def _disconnect_pending_overlay_refreshes(self):
        pending = list(self._pending_overlay_refresh_cids.items())
        self._pending_overlay_refresh_cids.clear()
        for marker, cid in pending:
            if cid is None:
                continue
            for viewer in self._iter_changed_viewers_for_color_update():
                canvas = getattr(viewer, "canvas", None)
                if canvas is None or id(canvas) != marker:
                    continue
                try:
                    canvas.mpl_disconnect(cid)
                except Exception:
                    pass
                break

    def _iter_colorbar_targets_for_viewer(self, viewer):
        targets = []
        seen = set()

        def _append(cax, colorbar, canvas):
            if cax is None or colorbar is None:
                return
            marker = id(cax)
            if marker in seen:
                return
            seen.add(marker)
            targets.append((cax, colorbar, canvas))

        get_state = getattr(viewer, "get_viewer_state", None)
        if callable(get_state):
            for plane in ("xy", "xz", "zy"):
                try:
                    state = get_state(plane)
                except Exception:
                    state = None
                if state is None:
                    continue
                _append(
                    getattr(state, "cax", None),
                    getattr(state, "colorbar", None),
                    getattr(state, "canvas", None),
                )

        _append(
            getattr(viewer, "cax", None),
            getattr(viewer, "colorbar", None),
            getattr(viewer, "canvas", None),
        )
        return targets

    def _reapply_colorbar_style_for_changed_viewers(self, config):
        viewers = self._iter_changed_viewers_for_color_update()
        for viewer in viewers:
            targets = self._iter_colorbar_targets_for_viewer(viewer)
            for cax, colorbar, canvas in targets:
                original_bounds = None
                try:
                    original_bounds = tuple(float(v) for v in cax.get_position().bounds)
                except Exception:
                    original_bounds = None

                try:
                    self.apply_colorbar_settings(cax, colorbar, config, caller=self)
                except Exception:
                    continue

                if original_bounds is not None:
                    try:
                        current_bounds = tuple(float(v) for v in cax.get_position().bounds)
                        if any(abs(cur - org) > 1e-9 for cur, org in zip(current_bounds, original_bounds)):
                            cax.set_position(list(original_bounds))
                    except Exception:
                        pass

                if canvas is not None:
                    try:
                        canvas.draw_idle()
                    except Exception:
                        pass

        for viewer in viewers:
            sig_fn = getattr(viewer, "_colorbar_layout_anchor_signature", None)
            if callable(sig_fn):
                try:
                    viewer._colorbar_auto_anchor_sig = sig_fn()
                except Exception:
                    pass

    def _defer_restore_colorbar_layout_after_pending_update(self):
        if bool(getattr(self, "_layout_restore_pending", False)):
            return
        self._layout_restore_pending = True

        def _run_restore():
            self._layout_restore_pending = False
            self._restore_colorbar_layout_after_pending_update()

        # Keep suppression for one UI frame so draw_idle callbacks triggered by
        # color updates do not re-schedule colorbar auto-layout.
        QTimer.singleShot(30, _run_restore)

    def _current_colorbar_config(self):
        viewer = getattr(self, "fits_viewer", None)
        config = None

        if self.mode == ColorMode.MAIN:
            config = getattr(getattr(viewer, "config_manager", None), "config", None)
        elif self.mode == ColorMode.CHANNEL:
            config = getattr(viewer, "config", None)
            if not isinstance(config, dict):
                root = getattr(viewer, "fits_viewer", None)
                config = getattr(getattr(root, "config_manager", None), "config", None)
        elif self.mode in (ColorMode.INTEG, ColorMode.PV):
            getter = getattr(viewer, "_get_colorbar_config", None)
            if callable(getter):
                try:
                    config = getter()
                except Exception:
                    config = None
            if not isinstance(config, dict):
                config = getattr(viewer, "config", None)
            if not isinstance(config, dict):
                root = getattr(viewer, "fits_viewer", None)
                config = getattr(getattr(root, "config_manager", None), "config", None)

        if not isinstance(config, dict):
            config = self.config if isinstance(self.config, dict) else {}
        self.config = config
        return config

    @staticmethod
    def _set_colorbar_layout_suppressed(viewer, suppressed):
        if viewer is None:
            return None
        previous = bool(getattr(viewer, "_suspend_colorbar_auto_layout", False))
        try:
            setattr(viewer, "_suspend_colorbar_auto_layout", bool(suppressed))
        except Exception:
            return None
        return previous

    def _refresh_overlay_after_color_change(self):
        for viewer in self._iter_changed_viewers_for_color_update():
            self._redraw_viewer_overlay_after_color_change(viewer)

    def _histogram_cache_owner(self):
        owner = getattr(self, "fits_viewer", None)
        return owner if owner is not None else self

    def _histogram_cache_key(self):
        raw_data = getattr(self, "data", None)
        array = np.asanyarray(raw_data)
        shape = tuple(int(axis) for axis in getattr(array, "shape", ()) or ())
        dtype = str(getattr(getattr(array, "dtype", None), "str", getattr(array, "dtype", "")) or "")
        nbytes = self._data_nbytes
        try:
            nbytes = int(nbytes) if nbytes is not None else None
        except Exception:
            nbytes = None
        return (
            str(getattr(self.mode, "value", self.mode)),
            id(raw_data),
            shape,
            dtype,
            int(getattr(array, "size", 0) or 0),
            nbytes,
        )

    def _get_cached_histogram(self):
        owner = self._histogram_cache_owner()
        cache = getattr(owner, "_color_panel_histogram_cache", None)
        if not isinstance(cache, dict):
            return None
        if cache.get("key") != self._histogram_cache_key():
            return None
        histogram = cache.get("histogram")
        return histogram if isinstance(histogram, dict) else None

    def _set_cached_histogram(self, histogram):
        owner = self._histogram_cache_owner()
        try:
            setattr(
                owner,
                "_color_panel_histogram_cache",
                {
                    "key": self._histogram_cache_key(),
                    "histogram": dict(histogram or {}),
                },
            )
        except Exception:
            pass

    def _compute_histogram_cache(self):
        data = np.asanyarray(self.data).ravel()
        if self._data_nbytes is not None and self._data_nbytes > MEMMAP_THRESHOLD_BYTES:
            stride = max(1, int(np.ceil(data.size / 1_000_000)))
            data = data[::stride]
        with np.errstate(invalid='ignore'):
            data = data[np.isfinite(data)]
        if data.size == 0:
            counts = np.zeros(100, dtype=float)
            bin_edges = np.linspace(0.0, 1.0, 101)
        else:
            counts, bin_edges = np.histogram(data, bins=100)
            counts = counts.astype(float, copy=False)
        return {
            "counts": counts,
            "bin_edges": bin_edges,
        }

    def _remove_histogram_artist(self):
        artist = getattr(self, "_histogram_artist", None)
        if artist is None:
            return
        try:
            if hasattr(artist, "remove"):
                artist.remove()
            elif hasattr(artist, "patches"):
                for patch in list(getattr(artist, "patches", []) or []):
                    try:
                        patch.remove()
                    except Exception:
                        pass
            elif isinstance(artist, (list, tuple)):
                for item in artist:
                    try:
                        item.remove()
                    except Exception:
                        pass
        except Exception:
            pass
        self._histogram_artist = None

    def update_histogram(self):
        histogram = self._get_cached_histogram()
        if histogram is None:
            histogram = self._compute_histogram_cache()
            self._set_cached_histogram(histogram)

        self._remove_histogram_artist()
        counts = np.asarray(histogram.get("counts", []), dtype=float)
        bin_edges = np.asarray(histogram.get("bin_edges", []), dtype=float)
        positive_counts = counts[counts > 0]

        if counts.size and bin_edges.size == counts.size + 1 and positive_counts.size:
            mask = counts > 0
            left_edges = bin_edges[:-1][mask]
            widths = np.diff(bin_edges)[mask]
            heights = counts[mask]
            self._histogram_artist = self.ax.bar(
                left_edges,
                heights,
                width=widths,
                align='edge',
                color=self.hist_color,
                edgecolor=self.hist_color,
                linewidth=0.0,
                alpha=1.0,
                zorder=1,
            )
            self.ax.set_xlim(float(bin_edges[0]), float(bin_edges[-1]))
            self.ax.set_yscale('log')
            self.ax.set_ylim(bottom=1.0)
        else:
            self.ax.set_yscale('linear')
            if bin_edges.size >= 2:
                self.ax.set_xlim(float(bin_edges[0]), float(bin_edges[-1]))
            self.ax.set_ylim(0.0, 1.0)

        self.fig.tight_layout()
        self.fig.subplots_adjust(bottom=0.2) 
        self.canvas.draw_idle()
    

    def init_vertical_lines(self):
        data_min, data_max = self._data_range()
        display_min, display_max = self._current_display_range()
        try:        
            if self.current_settings["min_val"] is not None and self.current_settings["max_val"] is not None:
                min_val = self.current_settings["min_val"]
                max_val = self.current_settings["max_val"]
            else:
                if self.intensity_min.text():
                    min_val = float(self.intensity_min.text())
                elif display_min is not None:
                    min_val = display_min
                else:
                    min_val = data_min

                if self.intensity_max.text():
                    max_val = float(self.intensity_max.text())
                elif display_max is not None:
                    max_val = display_max
                else:
                    max_val = data_max

        except ValueError:
            min_val = data_min
            max_val = data_max

        if min_val > max_val:
            min_val, max_val = max_val, min_val
        if min_val < data_min or min_val > data_max: min_val = data_min
        if max_val > data_max or max_val < data_min: max_val = data_max
        
        self.intensity_min.setText(str(f"{min_val:.3g}"))
        self.intensity_max.setText(str(f"{max_val:.3g}"))
        
        self.min_line = self.ax.axvline(min_val, color=self.vline_color, linestyle='-', linewidth=1, animated = False)
        self.max_line = self.ax.axvline(max_val, color=self.vline_color, linestyle='-', linewidth=1, animated = False)
        self.fig.tight_layout()
        self.canvas.draw_idle()
    
    def update_histogram_lines(self, min_val, max_val):
        self.min_line.set_xdata([min_val])
        self.max_line.set_xdata([max_val])
        self.canvas.draw_idle()
    
    def on_click(self, event):
        if event.inaxes != self.ax:
            return

        min_x = self.min_line.get_xdata()[0]
        max_x = self.max_line.get_xdata()[0]
        
        min_pixel_x = self.ax.transData.transform([min_x, 0])[0]
        max_pixel_x = self.ax.transData.transform([max_x, 0])[0]
        event_pixel_x = event.x
        if abs(event_pixel_x - min_pixel_x) < abs(event_pixel_x - max_pixel_x):
            if abs(event_pixel_x - min_pixel_x) < 15:  # 15 pixels threshold
                self.dragging_min = True
                self.min_line.set_animated(True)
        else:
            if abs(event_pixel_x - max_pixel_x) < 15:  # 15 pixels threshold
                self.dragging_max = True
                self.max_line.set_animated(True)
        self.canvas.draw()
        self.background = self.canvas.copy_from_bbox(self.ax.bbox)

        self.min_line.set_animated(False)
        self.max_line.set_animated(False)
        self.canvas.draw_idle()
        
    
    def on_motion(self, event):
        if event.inaxes != self.ax:
            return
        
        min_x = self.min_line.get_xdata()[0]
        max_x = self.max_line.get_xdata()[0]
        
        min_pixel_x = self.ax.transData.transform([min_x, 0])[0]
        max_pixel_x = self.ax.transData.transform([max_x, 0])[0]
        event_pixel_x = event.x
        
        if abs(event_pixel_x - min_pixel_x) < abs(event_pixel_x - max_pixel_x):
            if abs(event_pixel_x - min_pixel_x) <= 15:  # 15 pixels threshold
                self.canvas.setCursor(Qt.CursorShape.SizeHorCursor)
                self.min_line.set_linewidth(3)  # Increase line width
                self.max_line.set_linewidth(1)  # Reset line width
            else:
                self.canvas.setCursor(Qt.CursorShape.ArrowCursor)
                self.min_line.set_linewidth(1)  # Reset line width
                self.max_line.set_linewidth(1)  # Reset line width
            if event.button != 1: self.canvas.draw_idle()
        else:
            if abs(event_pixel_x - max_pixel_x) <= 15:  # 15 pixels threshold
                self.canvas.setCursor(Qt.CursorShape.SizeHorCursor)
                self.max_line.set_linewidth(3)  # Increase line width
                self.min_line.set_linewidth(1)  # Reset line width
            else:
                self.canvas.setCursor(Qt.CursorShape.ArrowCursor)
                self.min_line.set_linewidth(1)  # Reset line width
                self.max_line.set_linewidth(1)  # Reset line width
            if event.button != 1: self.canvas.draw_idle()
        

    
    def on_drag(self, event):
        if event.inaxes != self.ax or event.button != 1:
            return
            
        self.canvas.restore_region(self.background)
        
        if self.dragging_min:
            self.min_line.set_xdata([event.xdata])
            self.intensity_min.setText(str(f"{event.xdata:.3g}"))
            self.ax.draw_artist(self.min_line)
        elif self.dragging_max:
            self.max_line.set_xdata([event.xdata])
            self.intensity_max.setText(str(f"{event.xdata:.3g}"))
            self.ax.draw_artist(self.max_line)
        
        self.canvas.blit(self.ax.bbox)

    
    def on_release(self, event):
        if self.dragging_min or self.dragging_max:
            self.update_intensity_range()
        
        self.dragging_min = False
        self.dragging_max = False
        self.canvas.setCursor(Qt.CursorShape.ArrowCursor)
        self.min_line.set_linewidth(1)  # Reset line width
        self.max_line.set_linewidth(1)  # Reset line width
        self.min_line.set_animated(False)
        self.max_line.set_animated(False)


     
    def change_color_scale(self):
        pattern = self.colorscale_combo.currentText()
        if not pattern:
            QMessageBox.warning(self, 'Error', 'Color pattern cannot be empty.')
            return
        if pattern not in colormaps:
            QMessageBox.warning(self, 'Error', 'Invalid color pattern selected')
            self.colorscale_combo.setCurrentText(pattern.replace('_r', ''))  # Reset to last valid pattern
            return  # Exit the function if the pattern is invalid
        if self.invert_checkbox.isChecked():
            pattern += '_r'
            self.current_settings["invert"] = True
        #ColorSettingsPanel.color_pattern = pattern
        self.current_settings["color_pattern"] = pattern
        self.color_pattern = pattern

        try:
            def update_cmap(im):
                #im.set_cmap(pattern)
                im.set_cmap(self.create_gamma_cmap(pattern))
                im.cmap.set_bad(color=self.bad_color)
            self.apply_to_all_windows(update_cmap)
        except ValueError as ve:
            QMessageBox.warning(self, 'Error', str(ve))
            return
        self._schedule_shared_color_history("cmap")
        

    @classmethod
    def apply_colorbar_settings(cls, cax, colorbar, config, caller = None):
        if not isinstance(config, dict):
            config = {}

        tick_left = bool(config.get('colorbar_tick_left', False))
        tick_right = bool(config.get('colorbar_tick_right', True))
        tick_top = bool(config.get('colorbar_tick_top', False))
        tick_bottom = bool(config.get('colorbar_tick_bottom', True))
        label_left = bool(config.get('colorbar_tick_labelleft', False))
        label_right = not label_left
        label_top = bool(config.get('colorbar_tick_labeltop', False))
        label_bottom = not label_top
        tick_width = float(config.get('colorbar_tick_width', 1))
        tick_length = float(config.get('colorbar_tick_length', 4))
        mtick_length = float(config.get('colorbar_mtick_length', 2))
        tick_color = config.get('colorbar_tick_color')
        tick_labelcolor = config.get('colorbar_tick_labelcolor')
        tick_direction = config.get('colorbar_tick_direction')

        try:
            cax.set_zorder(300)
        except Exception:
            pass
        try:
            colorbar.ax.set_zorder(300)
        except Exception:
            pass

        cax.tick_params(
            axis='y',
            which='major',
            left=tick_left,
            right=tick_right,
            labelleft=label_left,
            labelright=label_right,
            width=tick_width,
            length=tick_length,
            color=tick_color,
            direction=tick_direction,
            labelcolor=tick_labelcolor,
        )
        cax.tick_params(
            axis='x',
            which='major',
            top=tick_top,
            bottom=tick_bottom,
            labeltop=label_top,
            labelbottom=label_bottom,
            width=tick_width,
            length=tick_length,
            color=tick_color,
            direction=tick_direction,
            labelcolor=tick_labelcolor,
        )
        cax.tick_params(axis='y', which='minor', left=tick_left, right=tick_right, length=mtick_length, color=tick_color, width=tick_width, direction=tick_direction)
        cax.tick_params(axis='x', which='minor', top=tick_top, bottom=tick_bottom, length=mtick_length, color=tick_color, width=tick_width, direction=tick_direction)

        try:
            if tick_left and tick_right:
                cax.yaxis.set_ticks_position('both')
            elif tick_left:
                cax.yaxis.set_ticks_position('left')
            elif tick_right:
                cax.yaxis.set_ticks_position('right')
            else:
                cax.yaxis.set_ticks_position('none')
        except Exception:
            pass
        try:
            if tick_top and tick_bottom:
                cax.xaxis.set_ticks_position('both')
            elif tick_top:
                cax.xaxis.set_ticks_position('top')
            elif tick_bottom:
                cax.xaxis.set_ticks_position('bottom')
            else:
                cax.xaxis.set_ticks_position('none')
        except Exception:
            pass
        try:
            cax.yaxis.set_label_position('left' if label_left else 'right')
        except Exception:
            pass
        try:
            cax.xaxis.set_label_position('top' if label_top else 'bottom')
        except Exception:
            pass

        colorbar.outline.set_color(tick_color)
        colorbar.outline.set_linewidth(tick_width)
        label_text = config.get('colorbar_label', '')
        if label_text is None:
            label_text = ''
        colorbar.set_label(str(label_text), fontsize=config.get('colorbar_label_fontsize'), color=config.get('colorbar_label_color'))
    
        colorbar.ax.minorticks_on()
        colorbar.ax.tick_params(which='minor', length=mtick_length, color=tick_color, width=tick_width, direction=tick_direction)
    
        if caller and hasattr(caller, 'log_checkbox') and caller.log_checkbox.isChecked():
            return
        try:
            mtick_freq = max(1, int(config.get('colorbar_mtick_freq', 2)))
        except Exception:
            mtick_freq = 2
        colorbar.ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(mtick_freq))
        colorbar.ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(mtick_freq))

    
    def create_gamma_cmap(self, pattern):
        cmap = plt.get_cmap(pattern)
        if pattern in ['Rainbow', 'Cool']:
            custom_cmap = CustomColormap(pattern, cmap._segmentdata)
            gamma_cmap = custom_cmap.apply_gamma(self.current_settings["gamma_value"])
            try:
                gamma_cmap.set_bad(color=self.config.get('bad_color', 'black'))
            except Exception:
                pass
            return gamma_cmap
        gamma_cmap = cmap(np.linspace(0, 1, cmap.N) ** self.current_settings["gamma_value"])
        listed = mpl.colors.ListedColormap(gamma_cmap)
        try:
            listed.set_bad(color=self.config.get('bad_color', 'black'))
        except Exception:
            pass
        return listed
    
    def update_gamma_from_spinbox(self):
        value = self.gamma_spinbox.value()
        self.current_settings["gamma_value"] = value
        self.apply_to_all_windows(lambda im: im.set_cmap(self.create_gamma_cmap(self.current_settings["color_pattern"])))
        self._schedule_shared_color_history("gamma")


    
    def closeEvent(self,event):
        self.current_settings["color_pattern"] = self.colorscale_combo.currentText()
        ColorSettingsPanel.settings[self.mode] = dict(self.current_settings)
        self._disconnect_pending_overlay_refreshes()
        try:
            plt.close(self.fig)
        except Exception:
            pass
        super().closeEvent(event)
